#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLDetector"""

import ast
import logging

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLDetector(AbstractDetector):
    """Detector class. Contains all information about detector"""

    def __init__(self, name):
        AbstractDetector.__init__(self, name)

        self.collect_name = None
        self.shutter_name = None
        self.temp_treshold = None
        self.hum_treshold = None
        self.cover_state = None
        self.binning_mode = None
        self.pixel_size_mm_x = None
        self.pixel_size_mm_y = None

        self.binding_mode = 1
        self.tolerance = None
        self.temperature = 0
        self.humidity = 0
        self.actual_frame_rate = None
        self.pixel_min = None
        self.pixel_max = None
        self.roi_modes_list = []
        self.roi_mode = None

        self.chan_beam_xy = None
        self.chan_cover_state = None
        self.chan_temperature = None
        self.chan_humidity = None
        self.chan_status = None
        self.chan_roi_mode = None
        self.chan_frame_rate = None
        self.chan_actual_frame_rate = None

        self.cmd_close_cover = None
        self.cmd_restart_daq = None

    def init(self):
        AbstractDetector.init(self)

        self.cover_state = "unknown"
        self.collect_name = self.get_property("collectName")
        self.shutter_name = self.get_property("shutterName")
        self.tolerance = self.get_property("tolerance")
        self.temp_treshold = self.get_property("tempThreshold")
        self.hum_treshold = self.get_property("humidityThreshold")
        self.pixel_min = self.get_property("px_min")
        self.pixel_max = self.get_property("px_max")
        self.roi_modes_list = ast.literal_eval(self.get_property("roiModes"))
        self.binning_mode = "Unbinned"

        self.pixel_size_mm_x = self.get_property("px")
        self.pixel_size_mm_y = self.get_property("py")

        self.chan_cover_state = self.get_channel_object("chanCoverState", optional=True)
        if self.chan_cover_state is not None:
            self.chan_cover_state.connect_signal("update", self.cover_state_changed)
        self.chan_temperature = self.get_channel_object("chanTemperature")
        self.chan_temperature.connect_signal("update", self.temperature_changed)
        self.chan_humidity = self.get_channel_object("chanHumidity")
        self.chan_humidity.connect_signal("update", self.humidity_changed)
        self.chan_status = self.get_channel_object("chanStatus")
        self.chan_status.connect_signal("update", self.status_changed)
        self.chan_roi_mode = self.get_channel_object("chanRoiMode")
        self.chan_roi_mode.connect_signal("update", self.roi_mode_changed)
        self.chan_frame_rate = self.get_channel_object("chanFrameRate")
        self.chan_frame_rate.connect_signal("update", self.frame_rate_changed)
        self.frame_rate_changed(self.chan_frame_rate.get_value())

        self.chan_actual_frame_rate = self.get_channel_object(
            "chanActualFrameRate", optional=True
        )
        if self.chan_actual_frame_rate is not None:
            self.chan_actual_frame_rate.connect_signal(
                "update", self.actual_frame_rate_changed
            )

        self.chan_beam_xy = self.get_channel_object("chanBeamXY")

        self.cmd_close_cover = self.get_command_object("cmdCloseCover")
        self.cmd_restart_daq = self.get_command_object("cmdRestartDaq")

    def has_shutterless(self):
        """Return True if has shutterless mode"""
        return self.get_property("hasShutterless")

    def get_collect_name(self):
        """Returns collection name"""
        return self.collect_name

    def get_shutter_name(self):
        """Returns shutter name"""
        return self.shutter_name

    def temperature_changed(self, value):
        """Updates temperature value"""
        if self.temperature is None or abs(self.temperature - value) > self.tolerance:
            self.temperature = value
            self.emit("temperatureChanged", (value, value < self.temp_treshold))
            self.status_changed("dummy")

    def humidity_changed(self, value):
        """Update humidity value"""
        if self.humidity is None or abs(self.humidity - value) > self.tolerance:
            self.humidity = value
            self.emit("humidityChanged", (value, value < self.hum_treshold))
            self.status_changed("dummy")

    def status_changed(self, status):
        """Status changed event"""
        status = "uninitialized"
        status_message = "Detector: "

        if self.chan_status is not None:
            status = self.chan_status.get_value()

        if self.temperature > self.temp_treshold:
            msg = "Detector: Temperature %0.2f is greater than allowed %0.2f" % (
                self.temperature,
                self.temp_treshold,
            )
            logging.getLogger("GUI").warning(msg)
            status_message = "Temperature has exceeded threshold.\n"

        if self.humidity > self.hum_treshold:
            msg = "Detector: Humidity %0.2f is greater than allowed %0.2f" % (
                self.humidity,
                self.hum_treshold,
            )
            logging.getLogger("GUI").warning(msg)
            status_message = status_message + "Humidity has exceeded threshold.\n"

        if status == "calibrating":
            status_message = status_message + "Calibrating. Please wait...\n"
            logging.getLogger("GUI").warning(status_message)

        self.emit("statusChanged", (status, status_message))

    def roi_mode_changed(self, mode):
        """ROI mode change event"""
        self.roi_mode = self.roi_modes_list.index(str(mode))
        self.emit("detectorRoiModeChanged", (self.roi_mode,))

    def frame_rate_changed(self, frame_rate):
        """Updates frame rate"""
        if frame_rate is not None:
            self._exposure_time_limits = (1 / float(frame_rate), 6000)

        self.emit("expTimeLimitsChanged", (self._exposure_time_limits,))

    def actual_frame_rate_changed(self, value):
        """Updates actual frame rate"""
        self.actual_frame_rate = value
        self.emit("frameRateChanged", value)

    def set_roi_mode(self, mode):
        """Sets roi mode

        :param mode: roi mode
        :type mode: str
        """
        self.chan_roi_mode.set_value(self.roi_modes_list[mode])

    def get_beam_centre(self, distance=None, wavelength=None):
        """Returns beam center coordinates"""
        beam_x = 0
        beam_y = 0
        if self.chan_beam_xy is not None:
            value = self.chan_beam_xy.get_value()
            beam_x = value[0]
            beam_y = value[1]
        return beam_x, beam_y

    def get_beam_centre_pix(self):
        beam_x, beam_y = self.get_beam_centre()
        return beam_x / 0.075, beam_y / 0.075

    def get_pixel_size_mm(self):
        """Returns pixel size in mm"""
        return self.pixel_size_mm_x, self.pixel_size_mm_y

    def cover_state_changed(self, state):
        """Updates guillotine state"

        :param state: guillotine state (close, opened, ..)
        :type state: str
        """
        if isinstance(state, (list, tuple)):
            state = state[1]

        if state == 0:
            self.cover_state = "closed"
        elif state == 1:
            self.cover_state = "opened"
        elif state == 2:
            self.cover_state = "closing"
        elif state == 3:
            self.cover_state = "opening"
        return self.cover_state

    def get_cover_state(self):
        """Returns cover state

        :return: str
        """
        return self.cover_state_changed(self.chan_cover_state.get_value())

    def is_cover_closed(self):
        """Returns True if cover is closed

        :return: bool
        """
        self.get_cover_state()
        return self.cover_state == "closed"

    def close_cover(self, wait=True):
        """Closes detector cover

        :param wait: wait
        :type wait: bool
        :return: None
        """
        if self.get_cover_state() != "closed":
            self.cmd_close_cover([0, 0])
            if wait:
                # with gevent.Timeout(15, Exception("Timeout waiting for detector cover
                # to close")):
                while self.get_cover_state() != "closed":
                    gevent.sleep(0.05)

    def restart_daq(self):
        """Restarts detector DAQ

        :return: None
        """
        self.cmd_restart_daq(0)

    def re_emit_values(self):
        """Reemits signals"""
        self.emit("detectorRoiModeChanged", (self.roi_mode,))
        temp = self.chan_temperature.get_value()
        self.emit("temperatureChanged", (temp, temp < self.temp_treshold))
        hum = self.chan_humidity.get_value()
        self.emit("humidityChanged", (hum, hum < self.hum_treshold))
        self.status_changed("")
        self.emit("expTimeLimitsChanged", (self._exposure_time_limits,))
        self.emit("frameRateChanged", self.actual_frame_rate)



#
#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""LNLS Energy"""

import logging

from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy
from mxcubecore.HardwareObjects.LNLS.EPICSActuator import EPICSActuator


class LNLSEnergy(EPICSActuator, AbstractEnergy):
    """LNLSEnergy class"""

    def init(self):
        """Initialise default properties"""
        super(LNLSEnergy, self).init()
        self.update_state(self.STATES.READY)
        self.detector = self.get_object_by_role("detector")

    def set_value(self):
        """Override method."""
        pass

    def get_value(self):
        """Read the actuator position.
        Returns:
            float: Actuator position.
        """
        value = super().get_value()
        # Nominal value stores last energy value with valid threshold energy
        # if abs(self._nominal_value - value) < 0.001:
        #    logging.getLogger("HWR").info("Pilatus threshold is still okay.")
        #    return value

        threshold_ok = self.check_threshold_energy(value)
        if threshold_ok:
            self._nominal_value = value
        else:
            value = None  # Invalid energy because threshold is invalid

        return value

    def check_threshold_energy(self, energy):
        """Returns whether detector threshold energy is valid or not."""

        # logging.getLogger("HWR").info(
        # "Checking Pilatus threshold. Please wait..."
        # )
        # for i in range(3):
        #    logging.getLogger("user_level_log").info(
        #        "Checking Pilatus threshold. Please wait..."
        #    )
        threshold_ok = self.detector.set_threshold_energy(energy)

        if threshold_ok:
            logging.getLogger("HWR").info("Pilatus threshold is okay.")
            # logging.getLogger("user_level_log").info(
            #    "Pilatus threshold is okay."
            # )
            return True

        logging.getLogger("HWR").error("Pilatus threshold is not okay.")
        # logging.getLogger("user_level_log").error(
        #    "Pilatus threshold is not okay."
        # )
        return False



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"

import base64
import copy
from functools import reduce
from io import BytesIO

import numpy as np
from PIL import Image

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractSampleView import (
    AbstractSampleView,
    ShapeState,
)
from mxcubecore.model import queue_model_objects


def combine_images(img1, img2):
    if img1.size != img2.size:
        raise ValueError("Images must be the same size")

    combined_img = Image.new("RGB", img1.size)

    pixels1 = img1.load()
    pixels2 = img2.load()
    combined_pixels = combined_img.load()

    width, height = img1.size
    for x in range(width):
        for y in range(height):
            pixel1 = pixels1[x, y]
            pixel2 = pixels2[x, y]

            if pixel2[0] <= 200 and pixel2[1] <= 60 and pixel2[2] <= 140:
                combined_pixels[x, y] = pixel1
            else:
                combined_pixels[x, y] = pixel2

    return combined_img


class SampleView(AbstractSampleView):
    def __init__(self, name):
        AbstractSampleView.__init__(self, name)
        self._shapes = {}

    def init(self):
        super(SampleView, self).init()
        self._last_oav_image = None

        self.hide_grid_threshold = self.get_property("hide_grid_threshold", 5)
        for motor_name, motor_ho in HWR.beamline.diffractometer.get_motors().items():
            if motor_ho:
                motor_ho.connect("stateChanged", self._update_shape_positions)

    def _update_shape_positions(self, *args, **kwargs):
        for shape in self.get_shapes():
            shape.update_position(HWR.beamline.diffractometer.motor_positions_to_screen)

        self.emit("shapesChanged")

    @property
    def shapes(self):
        return self._shapes

    def start_centring(self, tree_click=True):
        """
        Starts centring procedure
        """
        pass

    def cancel_centring(self):
        """
        Cancels current centring procedure
        """
        pass

    def start_auto_centring(self):
        """
        Start automatic centring procedure
        """
        pass

    def get_snapshot(self, overlay=None, bw=False, return_as_array=False):
        """
        Get snapshot(s)

        Args:
            overlay(str): Image data with shapes and other items to display on the snapshot
            bw(bool): return grayscale image
            return_as_array(bool): return as np array

        Returns:
            (BytesIO) snapshot as bytes image
        """
        img = self.take_snapshot(overlay_data=overlay, bw=bw)

        if return_as_array:
            return np.array(img)

        buffered = BytesIO()
        img.save(buffered, format="JPEG")

        return buffered

    def save_snapshot(self, path, overlay=None, bw=False):
        """
        Save a snapshot to file.

        Args:
            path (str): The filename.
            overlay(str): Image data with shapes and other items to display on the snapshot
            bw(bool): return grayscale image
        """
        img = self.take_snapshot(overlay_data=overlay, bw=bw)
        img.save(path)

        self._last_oav_image = path

    def take_snapshot(self, overlay_data=None, bw=False):
        """
        Get snapshot with overlaid data.

        Args:
            overlay_data (str): base64 encoded image to lay over camera image
            bw (bool): return grayscale image

        Returns:
            (Image) rgb or grayscale image
        """
        data, width, height = self.camera.get_last_image()

        img = Image.frombytes("RGB", (width, height), data)

        if overlay_data:
            overlay_data = base64.b64decode(overlay_data)
            overlay_image = Image.open(BytesIO(overlay_data))
            overlay_image = overlay_image.resize(
                (width, height), Image.Resampling.LANCZOS
            )
            img = combine_images(img, overlay_image.convert("RGB"))

        if bw:
            img.convert("1")

        return img

    def get_last_image_path(self):
        return self._last_oav_image

    def add_shape(self, shape):
        """
        Add the shape <shape> to the dictionary of handled shapes.

        Args:
            param (shape): Shape to add.
            type (shape): Shape object.
        """
        self.shapes[shape.id] = shape
        shape.shapes_hw_object = self

    def add_shape_from_mpos(
        self,
        mpos_list,
        screen_coord,
        t,
        state: ShapeState = "SAVED",
        user_state: ShapeState = "SAVED",
    ):
        """
        Adds a shape of type <t>, with motor positions from mpos_list and
        screen position screen_coord.

        Args:
            mpos_list (list[mpos_list]): List of motor positions
            screen_coord (tuple(x, y): Screen coordinate for shape
            t (str): Type str for shape, P (Point), L (Line), G (Grid)

        Returns:
            (Shape) Shape of type <t>
        """
        cls_dict = {"P": Point, "L": Line, "G": Grid, "2DP": TwoDPoint}
        _cls = cls_dict[t]
        shape = None

        if _cls:
            shape = _cls(mpos_list, screen_coord)
            # In case the shape is being recreated, we need to restore it's state.
            shape.state = state
            shape.user_state = user_state

            self.add_shape(shape)

        return shape

    def add_shape_from_refs(
        self, refs, t, state: ShapeState = "SAVED", user_state: ShapeState = "SAVED"
    ):
        """
        Adds a shape of type <t>, taking motor positions and screen positions
        from reference points in refs.

        Args:
            refs (list[str]): List of id's of the reference Points
            t (str): Type str for shape, P (Point), L (Line), G (Grid)

        Returns:
            (Shape): Shape of type <t>
        """
        mpos = [self.get_shape(refid).mpos() for refid in refs]
        spos_list = [self.get_shape(refid).screen_coord for refid in refs]
        spos = reduce((lambda x, y: tuple(x) + tuple(y)), spos_list, ())
        shape = self.add_shape_from_mpos(mpos, spos, t, state, user_state)
        shape.refs = refs

        return shape

    def delete_shape(self, sid):
        """
        Removes the shape with id <sid> from the list of handled shapes.

        Args:
            sid (str): The id of the shape to remove

        Returns:
            (Shape): The removed shape
        """
        shape = self.shapes.pop(sid, None)

        if shape:
            shape.shapes_hw_object = None

        return shape

    def select_shape(self, sid):
        """
        Select the shape <shape>.

        Args:
            sid (str): Id of the shape to select.
        """
        shape = self.shapes.get(sid, None)

        if shape:
            shape.select()

    def de_select_shape(self, sid):
        """
        De-select the shape with id <sid>.

        Args:
            sid (str): The id of the shape to de-select.
        """
        shape = self.shapes.get(sid, None)

        if shape:
            shape.de_select()

    def is_selected(self, sid):
        """
        Check if Shape with <sid> is selected.

        Returns:
            (Boolean) True if Shape with <sid> is selected False otherwise
        """
        shape = self.shapes.get(sid, None)
        return bool(shape and shape.is_selected())

    def get_selected_shapes(self):
        """
        Get all selected shapes.

        Returns:
           (list[Shape]) List of selected Shapes
        """
        return [s for s in self.shapes.values() if s.is_selected()]

    def de_select_all(self):
        """De select all shapes."""

        for shape in self.shapes.values():
            shape.de_select()

    def select_shape_with_cpos(self, cpos):
        """
        Selects shape with the assocaitaed centered position <cpos>

        Args:
            cpos (CenteredPosition)
        """
        return

    def clear_all(self):
        """
        Clear the shapes, remove all contents.
        """
        self._shapes = {}
        Grid.SHAPE_COUNT = 0
        Line.SHAPE_COUNT = 0
        Point.SHAPE_COUNT = 0

    def get_shapes(self):
        """
        Get all Shapes.

        Returns:
            (list[Shape]) All the shapes
        """
        return self.shapes.values()

    def get_points(self):
        """
        Get all Points currently handled.

        Returns:
            (list[Point]) All points currently handled
        """
        current_points = []

        for shape in self.get_shapes():
            if isinstance(shape, Point):
                current_points.append(shape)

        return current_points

    def get_lines(self):
        """
        Get all Lines currently handled.

        Returns:
            (list[Line]) All lines currently handled
        """
        lines = []

        for shape in self.get_shapes():
            if isinstance(shape, Line):
                lines.append(shape)

        return lines

    def get_grids(self):
        """
        Get all Grids currently handled.

        Returns:
            (list[Grid]) All lines currently handled
        """
        grid = []

        for shape in self.get_shapes():
            if isinstance(shape, Grid):
                grid.append(shape)

        return grid

    def get_shape(self, sid):
        """
        Get Shape with id <sid>.

        Args:
            sid (str): id of Shape to retrieve

        Returns:
            (Shape) All the shapes
        """
        return self.shapes.get(sid, None)

    # For backwards compatibility with old ShapeHisotry object
    # returns first of selected grids
    def get_grid(self):
        """
        Get the first of the selected grids, (the one that was selected first in
        a sequence of select operations)

        Returns:
            (dict): The first selected grid as a dictionary
        """
        grid = None

        for shape in self.get_shapes():
            if isinstance(shape, Grid):
                grid = shape.as_dict()
                break

        return grid

    def set_grid_data(self, sid, result_data, data_file_path):
        """
        Sets grid rsult data for a shape with the specified id.

        Args:
            sid (str): The id of the shape to set grid data for.
            result_data: The result data to set for the shape. Either a base64 encoded string for PNG/image
            or a dictionary for RGB (keys are cell number and value RGBa list). Data is only updated if result is RGB based
            data_file_path (str): The path to the data file associated with the result data.

        Returns:
            None

        Raises:
            AttributeError: If no shape with the specified id exists.
        """

        shape = self.get_shape(sid)

        if shape:
            if shape.result and type(shape.result) == dict:
                # append data
                shape.result.update(result_data)
            else:
                shape.set_result(result_data)
                shape.result_data_path = data_file_path

            self.emit("newGridResult", shape)
        else:
            msg = "Cant set result for %s, no shape with id %s" % (sid, sid)
            raise AttributeError(msg)

    def get_grid_data(self, key):
        result = {}
        shape = self.get_shape(key)

        if shape:
            result = shape.get_result()

        return result

    def inc_used_for_collection(self, cpos):
        """
        Increase counter that keepts on collect made on this shape,
        shape with associated CenteredPosition cpos

        Args:
            cpos (CenteredPosition): CenteredPosition of shape
        """
        pass


class Shape(object):
    """
    Base class for shapes.
    """

    SHAPE_COUNT = 0

    def __init__(self, mpos_list=[], screen_coord=(-1, -1)):
        object.__init__(self)
        Shape.SHAPE_COUNT += 1
        self.t = "S"
        self.id = ""
        self.cp_list = []
        self.name = ""
        self.state: ShapeState = "SAVED"
        self.user_state: ShapeState = "SAVED"  # used to persist user preferences in regards whether to show or hide particular shape.
        self.label = ""
        self.screen_coord = screen_coord
        self.selected = False
        self.refs = []
        self.shapes_hw_object = None

        self.add_cp_from_mp(mpos_list)

    def get_centred_positions(self):
        """
        :returns: The centred position(s) associated with the shape.
        :rtype: List of CentredPosition objects.
        """
        return self.cp_list

    def get_centred_position(self):
        return self.get_centred_positions()[0]

    def select(self):
        self.selected = True

    def de_select(self):
        self.selected = False

    def is_selected(self):
        return self.selected

    def update_position(self, transform):
        spos_list = [transform(cp.as_dict()) for cp in self.cp_list]
        spos_list = tuple([pos for l in spos_list for pos in l])
        self.screen_coord = spos_list

    def add_cp_from_mp(self, mpos_list):
        for mp in mpos_list:
            self.cp_list.append(queue_model_objects.CentredPosition(mp))

    def set_id(self, id_num):
        self.id = self.t + "%s" % id_num
        self.name = self.label + "-%s" % id_num

    def move_to_mpos(self, mpos_list, screen_coord=[]):
        self.cp_list = []
        self.add_cp_from_mp(mpos_list)

        if screen_coord:
            self.screen_coord = screen_coord

    def update_from_dict(self, shape_dict):
        # We don't allow id or result updates
        shape_dict.pop("id", None)
        shape_dict.pop("result", None)

        for key, value in shape_dict.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def as_dict(self):
        cpos_list = []

        for cpos in self.cp_list:
            cpos_list.append(cpos.as_dict())

        d = copy.deepcopy(vars(self))

        # Do not serialize Shapes HW Object
        d.pop("shapes_hw_object")

        # replace cpos_list with a list of motor positions
        d.pop("cp_list")
        d["motor_positions"] = str(cpos_list)

        return d


class Point(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Shape.__init__(self, mpos_list, screen_coord)
        Point.SHAPE_COUNT += 1
        self.t = "P"
        self.label = "Point"
        self.set_id(Point.SHAPE_COUNT)

    def mpos(self):
        return self.cp_list[0].as_dict()

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def as_dict(self):
        d = Shape.as_dict(self)
        # replace cpos_list with the motor positions
        d["motor_positions"] = self.cp_list[0].as_dict()
        return d


class TwoDPoint(Point):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Point.__init__(self, mpos_list, screen_coord)
        self.t = "2DP"
        self.label = "2D-Point"
        self.set_id(Point.SHAPE_COUNT)


class Line(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Shape.__init__(self, mpos_list, screen_coord)
        Line.SHAPE_COUNT += 1
        self.t = "L"
        self.label = "Line"
        self.set_id(Line.SHAPE_COUNT)

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def get_centred_positions(self):
        return [self.start_cpos, self.end_cpos]

    def get_points_index(self):
        if all(self.cp_list):
            return (self.cp_list[0].get_index(), self.cp_list[1].get_index())


class Grid(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Shape.__init__(self, mpos_list, screen_coord)
        Grid.SHAPE_COUNT += 1
        self.t = "G"
        self.set_id(Grid.SHAPE_COUNT)

        self.width = -1
        self.height = -1
        self.cell_count_fun = "zig-zag"
        self.cell_h_space = -1
        self.cell_height = -1
        self.cell_v_space = -1
        self.cell_width = -1
        self.label = "Grid"
        self.num_cols = -1
        self.num_rows = -1
        self.selected = False
        # result is a base64 encoded string for PNG/image heatmap results
        # or a dictionary (for RGB number based results)
        self.result = None
        self.pixels_per_mm = [1, 1]
        self.beam_pos = [1, 1]
        self.beam_width = 0
        self.beam_height = 0
        self.hide_threshold = 5

        self.set_id(Grid.SHAPE_COUNT)

    def update_position(self, transform):
        phi_pos = HWR.beamline.diffractometer.omega.get_value() % 360
        _d = abs((self.get_centred_position().phi % 360) - phi_pos)

        if self.user_state == "HIDDEN":
            self.state = "HIDDEN"
            return

        if min(_d, 360 - _d) > self.shapes_hw_object.hide_grid_threshold:
            self.state = "HIDDEN"
        else:
            super(Grid, self).update_position(transform)
            self.state = "SAVED"

    def get_centred_position(self):
        return self.cp_list[0]

    def get_grid_range(self):
        return (
            float(self.cell_width * (self.num_cols - 1)),
            float(self.cell_height * (self.num_rows - 1)),
        )

    def get_num_lines(self):
        if self.cell_count_fun == "zig-zag":
            return self.num_rows
        elif self.cell_count_fun == "inverse-zig-zag":
            return self.num_cols
        else:
            return self.num_rows

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def set_result(self, result_data):
        self.result = result_data

    def get_result(self):
        return self.result

    def as_dict(self):
        d = Shape.as_dict(self)
        # replace cpos_list with the motor positions
        d["motor_positions"] = self.cp_list[0].as_dict()

        pixels_per_mm = HWR.beamline.diffractometer.get_pixels_per_mm()
        beam_pos = HWR.beamline.beam.get_beam_position_on_screen()
        size_x, size_y, shape, _label = HWR.beamline.beam.get_value()

        d["x1"] = -float((beam_pos[0] - d["screen_coord"][0]) / pixels_per_mm[0])
        d["y1"] = -float((beam_pos[1] - d["screen_coord"][1]) / pixels_per_mm[1])
        d["steps_x"] = d["num_cols"]
        d["steps_y"] = d["num_rows"]
        d["dx_mm"] = d["width"] / pixels_per_mm[0]
        d["dy_mm"] = d["height"] / pixels_per_mm[1]
        d["beam_width"] = size_x
        d["beam_height"] = size_y
        d["angle"] = 0

        return d



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""

[Name] SOLEILMachineInfo
based on EMBLMachineInfo

[Description]
Hardware Object is used to get relevant machine information
(current, intensity, hutch temperature and humidity, and data storage disc
information). Value limits are included

[Channels]
- chanMachCurr
- chanStateText

[Commands]
- cmdSetIntensResolution
- cmdSetIntensAcqTime
- cmdSetIntensRange

[Emitted signals]
- valuesChanged
- inRangeChanged

[Functions]
- mach_current_changed()
- machStateTextChanged()
- updateValues()
- setInitialIntens()
- setExternalValues()


[Included Hardware Objects]

Example Hardware Object XML file :
==================================
<object class="MachineInfo">
    <updateIntervalS>120</updateIntervalS>
    <discPath>/home</discPath>
    <limits>{'current':90, 'temp': 25, 'hum': 60, 'intens': 0.1,
             'discSizeGB': 20}</limits>
</object>
"""

import logging
import os
import time
from datetime import (
    datetime,
    timedelta,
)

from gevent import spawn
from urllib2 import urlopen

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["SOLEIL", "EMBL Hamburg"]
__version__ = "2.3."
__category__ = "General"


class SOLEILMachineInfo(HardwareObject):
    """
    Descript. : Displays actual information about the beeamline
    """

    def __init__(self, name):
        """__init__"""
        HardwareObject.__init__(self, name)
        # Parameters
        self.update_interval = None
        self.limits_dict = None
        self.hutch_temp_addr = None
        self.hutch_hum_addr = None
        self.hutch_temp = 0
        self.hutch_hum = 0
        self.overflow_alarm = None
        self.low_level_alarm = 0
        self.auto_refill = None
        self.state_text = "Not updated yet"
        self.ring_energy = None
        self.bunch_count = None
        self.flux_area = None
        self.last_transmission = None

        self.values_list = []
        # Intensity current ranges
        # Machine current, index = 0
        temp_dict = {}
        temp_dict["value"] = 0
        temp_dict["value_str"] = ""
        temp_dict["in_range"] = False
        temp_dict["title"] = "Machine current"
        temp_dict["bold"] = True
        temp_dict["font"] = 14
        # temp_dict['history'] = True
        self.values_list.append(temp_dict)

        # Machine state, index = 1
        temp_dict = {}
        temp_dict["value"] = None
        temp_dict["in_range"] = True
        temp_dict["title"] = "Machine state"
        self.values_list.append(temp_dict)

        # Hutch temperature, index = 2
        temp_dict = {}
        temp_dict["value"] = ""
        temp_dict["value_str"] = ""
        temp_dict["in_range"] = None
        temp_dict["title"] = "Hutch temperature"
        self.values_list.append(temp_dict)

        # Remeasure flux, index = 3
        temp_dict = {}
        temp_dict["value"] = 1
        temp_dict["value_str"] = "Remeasure flux!"
        temp_dict["in_range"] = False
        temp_dict["title"] = "Flux"
        temp_dict["align"] = "left"
        self.values_list.append(temp_dict)

        # Cryo jet, index = 4
        temp_dict = {}
        temp_dict["value"] = "???"
        temp_dict["in_range"] = None
        temp_dict["title"] = "Cryostream"
        self.values_list.append(temp_dict)

        # Dewar level, index = 5
        temp_dict = {}
        temp_dict["value"] = "Dewar level in range"
        temp_dict["in_range"] = True
        temp_dict["title"] = "Sample changer"
        self.values_list.append(temp_dict)

        self.temp_hum_values = [None, None]
        self.temp_hum_in_range = [None, None]
        self.temp_hum_polling = None

        self.chan_mach_curr = None
        self.chan_mach_energy = None
        self.chan_bunch_count = None
        self.chan_state_text = None
        self.chan_cryojet_in = None
        self.chan_sample_temperature = None
        self.chan_sc_dewar_low_level_alarm = None
        self.chan_sc_dewar_overflow_alarm = None

        self.ring_energy = 2.75

    def init(self):
        """init"""
        self.update_interval = int(self.get_property("updateIntervalS"))
        self.limits_dict = eval(self.get_property("limits"))

        self.chan_mach_curr = self.get_channel_object("machCurrent")
        if self.chan_mach_curr is not None:
            self.chan_mach_curr.connect_signal("update", self.mach_current_changed)

        self.chan_filling_mode = self.get_channel_object("fillingMode")
        if self.chan_filling_mode is not None:
            self.chan_filling_mode.connect_signal("update", self.state_text_changed)

        self.chan_state_text0 = self.get_channel_object("operatorMessage0")
        if self.chan_state_text0 is not None:
            self.chan_state_text0.connect_signal("update", self.state_text_changed)

        self.chan_state_text1 = self.get_channel_object("operatorMessage1")
        if self.chan_state_text1 is not None:
            self.chan_state_text1.connect_signal("update", self.state_text_changed)

        self.chan_state_text2 = self.get_channel_object("operatorMessage2")
        if self.chan_state_text2 is not None:
            self.chan_state_text2.connect_signal("update", self.state_text_changed)

        self.chan_is_beam_usable = self.get_channel_object("isBeamUsable")
        if self.chan_is_beam_usable is not None:
            self.chan_is_beam_usable.connect_signal("update", self.state_text_changed)

        self.chan_cryojet_in = self.get_channel_object("cryojetIn")
        if self.chan_cryojet_in is not None:
            self.cryojet_in_changed(self.chan_cryojet_in.get_value())
            self.chan_cryojet_in.connect_signal("update", self.cryojet_in_changed)
        else:
            logging.getLogger("HWR").debug("MachineInfo: Cryojet channel not defined")

        self.chan_sample_temperature = self.get_channel_object("sampleTemp")
        if self.chan_sample_temperature is not None:
            # self.chan_sample_temperature.connect_signal('update', self.cryojet_in_changed)
            self.cryojet_in_changed(self.chan_cryojet_in.get_value())

        self.chan_sc_auto_refill = self.get_channel_object("scAutoRefill")
        if self.chan_sc_auto_refill is not None:
            self.chan_sc_auto_refill.connect_signal(
                "update", self.sc_autorefill_changed
            )
            self.sc_autorefill_changed(self.chan_sc_auto_refill.get_value())

        self.chan_sc_dewar_low_level_alarm = self.get_channel_object("scLowLevelAlarm")
        if self.chan_sc_dewar_low_level_alarm is not None:
            self.chan_sc_dewar_low_level_alarm.connect_signal(
                "update", self.low_level_alarm_changed
            )
            self.low_level_alarm_changed(self.chan_sc_dewar_low_level_alarm.get_value())

        self.chan_sc_dewar_overflow_alarm = self.get_channel_object("scOverflowAlarm")
        if self.chan_sc_dewar_overflow_alarm is not None:
            self.chan_sc_dewar_overflow_alarm.connect_signal(
                "update", self.overflow_alarm_changed
            )

        # self.chan_flux = self.get_channel_object('flux')
        # if self.chan_flux is not None:
        # self.chan_flux.connect_signal('update', self.flux_changed)

        self.chan_temperature_exp = self.get_channel_object("temperatureExp")
        if self.chan_temperature_exp is not None:
            self.chan_temperature_exp.connect_signal("update", self.temperature_changed)
            self.temperature_changed(self.chan_temperature_exp.get_value())

        self.re_emit_values()

    def clear_gevent(self):
        self.temp_hum_polling.kill()
        if self.update_task:
            self.update_task.kill()

    def cryojet_in_changed(self, value):
        """Cryojet in/out value changed"""
        logging.getLogger("HWR").debug("cryojet_in_changed: %s" % value)
        self.values_list[4]["in_range"] = False
        self.values_list[4]["bold"] = True

        if value == 0:
            self.values_list[4]["value"] = " In place"
            self.values_list[4]["in_range"] = True
            self.values_list[4]["bold"] = False
        elif value == 1:
            self.values_list[4]["value"] = "NOT IN PLACE"
        else:
            self.values_list[4]["value"] = "Unknown"

        if self.chan_sample_temperature is not None:
            self.values_list[4]["value"] += (
                "\n sample temperature: %.1f K"
                % self.chan_sample_temperature.get_value()
            )
        else:
            logging.getLogger("HWR").debug(
                "chan_sample_temperature: %s" % self.chan_sample_temperature
            )
        self.re_emit_values()

    def mach_current_changed(self, value):
        """Method called if the machine current is changed

        :param value: new machine current
        :type value: float
        """
        if (
            self.values_list[0]["value"] is None
            or abs(self.values_list[0]["value"] - value) > 0.00001
        ):
            self.values_list[0]["value"] = value
            self.values_list[0]["value_str"] = "%.1f mA" % value
            self.values_list[0]["in_range"] = value > 60.0
            self.re_emit_values()

    def state_text_changed(self, text):
        """Function called if machine state text is changed

        :param text: new machine state text
        :type text: string
        """
        # self.state_text = str(text)
        # self.values_list[1]['in_range'] = text != "Fehler"
        self.update_machine_state()

    def update_machine_state(self):
        """Machine state assembly"""
        filling_mode = self.chan_filling_mode.get_value()
        state_text0 = self.chan_state_text0.get_value()
        state_text1 = self.chan_state_text1.get_value()
        state_text2 = self.chan_state_text2.get_value()
        is_beam_usable = self.chan_is_beam_usable.get_value()

        date_boundary_string = " :"
        date_boundary = state_text1.find(date_boundary_string)
        date = state_text1[:date_boundary]
        state_text1 = state_text1[date_boundary + len(date_boundary_string) :]
        state_text = "%s, %s\n" % (date, state_text0)
        state_text += "electron energy: %.2f GeV, filling: %s\n" % (
            self.ring_energy,
            filling_mode,
        )
        state_text += "%s\n" % (state_text1,)
        if state_text2 != " ":
            state_text += "%s\n" % state_text2

        if is_beam_usable:
            self.values_list[1]["in_range"] = True
            state_text += "Beam usable"
        else:
            self.values_list[1]["in_range"] = False
            state_text += "Beam unusable"
        self.values_list[1]["value"] = state_text
        self.state_text = state_text
        self.re_emit_values()

    def low_level_alarm_changed(self, value):
        """Low level alarm"""
        self.low_level_alarm = value
        self.update_sc_alarm()

    def overflow_alarm_changed(self, value):
        """Overflow alarm"""
        self.overflow_alarm = value
        self.update_sc_alarm()

    def sc_autorefill_changed(self, value):
        self.auto_refill = value
        self.update_sc_alarm()

    def file_transfer_status_changed(self, total, pending, failed):
        self.values_list[-1]["value"] = "%d  -  %d  -  %d" % (total, pending, failed)
        self.values_list[-1]["in_range"] = failed == 0

        if failed > 0:
            logging.getLogger("GUI").error(
                "Error in file transfer (%d files failed to copy)." % failed
            )

    def update_sc_alarm(self):
        """Sample changer alarm"""
        if self.low_level_alarm == 1:
            self.values_list[5]["value"] = "Low level alarm!"
            self.values_list[5]["in_range"] = False
            self.values_list[5]["bold"] = True
            # logging.getLogger("GUI").error("Liquid nitrogen " + \
            # " level in sample changer dewar is too low!")

        elif self.overflow_alarm:
            self.values_list[5]["value"] = "Overflow alarm!"
            self.values_list[5]["in_range"] = False
            self.values_list[5]["bold"] = True
            logging.getLogger("GUI").error(
                "Liquid nitrogen " + "overflow in sample changer dewar!"
            )
        else:
            self.values_list[5]["value"] = "Dewar level in range"
            self.values_list[5]["in_range"] = True

        logging.getLogger("HWR").error(
            "chan_sc_auto_refill %s" % self.chan_sc_auto_refill.get_value()
        )
        if self.chan_sc_auto_refill.get_value() == 0:
            self.values_list[5]["value"] += ", refill OFF"
        else:
            self.values_list[5]["value"] += ", refill ON"
        self.re_emit_values()

    def flux_changed(self, value, beam_info=None, transmission=None):
        """Sets flux value"""
        if value is None:
            value = -1
        self.values_list[3]["value"] = value
        msg_str = "Flux: %.2E ph/s" % value
        # msg_str += "\n@ %.1f transmission , %d x %d beam" % (\
        # transmission, beam_info['size_x'] * 1000, beam_info['size_y'] * 1000)
        self.values_list[3]["value_str"] = msg_str
        self.values_list[3]["in_range"] = value > 1e6
        self.re_emit_values()

    def re_emit_values(self):
        """Emits list of values"""
        self.emit("valuesChanged", self.values_list)

    def get_values(self):
        """Returns list of values"""
        val = dict(self.values_list)
        return val

    def temperature_changed(self, value):
        """ "Update hutch temperature"""
        self.values_list[2]["value"] = "%.1f C" % value
        self.values_list[2]["in_range"] = value < 25  # self.limits_dict['temp']
        self.re_emit_values()

    def get_temp_hum_values(self, sleep_time):
        """Updates temperature and humidity values"""
        while True:
            temp = self.get_external_value(self.hutch_temp_addr)
            hum = self.get_external_value(self.hutch_hum_addr)
            if not None in (temp, hum):
                if abs(float(temp) - self.hutch_temp) > 0.1 or abs(
                    float(hum) != self.hutch_hum > 1
                ):
                    self.hutch_temp = temp
                    self.hutch_hum = hum
                    self.values_list[2]["value"] = "%.1f C, %.1f %%" % (temp, hum)
                    self.values_list[2]["in_range"] = temp < 25 and hum < 60
                    self.re_emit_values()
            time.sleep(sleep_time)

    def get_current(self):
        "Returns current"
        return self.values_list[0]["value"]

    def get_current_value(self):
        """Returns current"""
        return self.values_list[0]["value"]

    def get_message(self):
        """Returns synchrotron state text"""
        return self.state_text

    def update_ramdisk_size(self, sleep_time):
        while True:
            total, free, perc = self.get_ramdisk_size()
            if None in (total, free, perc):
                txt = " Unable to read ramdisk size!"
                self.values_list[-1]["in_range"] = False
            else:
                txt = " Total: %s\n Free:  %s (%s)" % (
                    self.sizeof_fmt(total),
                    self.sizeof_fmt(free),
                    "{0:.0%}".format(perc),
                )
                self.values_list[-1]["in_range"] = free / 2**30 > 10
            self.values_list[-1]["value"] = txt
            self.re_emit_values()
            time.sleep(sleep_time)

    def get_ramdisk_size(self):
        data_dir = "/ramdisk/"
        if os.path.exists(data_dir):
            st = os.statvfs(data_dir)

            total = st.f_blocks * st.f_frsize
            free = st.f_bavail * st.f_frsize
            perc = st.f_bavail / float(st.f_blocks)
            return total, free, perc
        else:
            return None, None, None

    def sizeof_fmt(self, num):
        """Returns disk space formatted in string"""

        try:
            for x in ["bytes", "KB", "MB", "GB"]:
                if num < 1024.0:
                    return "%3.1f%s" % (num, x)
                num /= 1024.0
            return "%3.1f%s" % (num, "TB")
        except Exception:
            return "???"



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import numpy as np
from scipy.interpolate import (
    Rbf,
    interp1d,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractFlux import AbstractFlux

__credits__ = ["MXCuBE collaboration"]
__category__ = "General"


class P11Flux(AbstractFlux):
    def __init__(self, name):
        super().__init__(name)

        self.measured_flux_list = []
        self.measured_flux_dict = {}
        self.current_flux_dict = {}

    def init(self):
        # Here are reference flux values from the config file
        # Those are values at which flux is measured
        self.default_flux = self.get_property("defaultFlux")

        self.default_beamsize = self.get_property("defaultBeamsize")
        self.default_pinhole = self.get_property("defaultPinhole")
        self.default_energy = self.get_property("defaultEnergy")
        self.default_current = self.get_property("defaultCurrent")
        self.measure_flux()

    def get_value(self):
        """Get flux at current transmission in units of photons/s"""
        # TODO: Once the motor movements are involved, check the logic.
        self.measure_flux()
        return self.current_flux_dict["flux"]

    def measure_flux(self):
        """Measures intensity"""
        beam_size = HWR.beamline.beam.get_beam_size()
        transmission = HWR.beamline.transmission.get_value()
        energy = HWR.beamline.energy.get_value() * 1000
        # TODO: Check pinhole from HWR.beamline.diffractometer
        current_pinhole = HWR.beamline.beam.get_pinhole_size()
        current = HWR.beamline.machine_info.get_current()
        mess = HWR.beamline.machine_info.get_message()

        flux = (
            self.estimate_flux_with_reference(
                max(beam_size[0] * 1000, beam_size[0] * 1000),
                max(beam_size[0] * 1000, beam_size[0] * 1000),
                energy,
                current,
                self.default_beamsize,
                self.default_pinhole,
                self.default_energy,
                self.default_flux,
                self.default_current,
            )
            * transmission
            / 100.0
        )

        self.measured_flux_list = [
            {
                "size_x": beam_size[0],
                "size_y": beam_size[1],
                "transmission": transmission,
                "flux": flux,
            }
        ]

        self.measured_flux_dict = self.measured_flux_list[0]
        self.current_flux_dict = self.measured_flux_list[0]

        self.emit(
            "fluxInfoChanged",
            {"measured": self.measured_flux_dict, "current": self.current_flux_dict},
        )

        self.log.debug(
            f"Estimated flux for beam size {max(beam_size[0] * 1000, beam_size[0] * 1000)}, pinhole size {current_pinhole}, transmission {transmission}, energy {energy} eV and current {current} mA: {flux:.2e} ph/s"
        )

    def estimate_flux_with_reference(
        self,
        beam_size,
        pinhole_size,
        energy,
        current,
        ref_beam_size,
        ref_pinhole_size,
        ref_energy,
        ref_flux,
        ref_current,
    ):
        """
        Helper function to estimate the flux based on the input beam size, pinhole size, energy, and current, scaled by a reference flux and current.
        Temporary plug to have realistic flux estimations.

        Parameters:
        beam_size (int): Beam size in numerical format (e.g., 300, 200, 100, 50, 20, 9)
        pinhole_size (int): Pinhole size in numerical format (e.g., 200, 100, 50, 20)
        energy (int): Energy in eV
        current (float): Current in mA
        ref_beam_size (int): Reference beam size
        ref_pinhole_size (int): Reference pinhole size
        ref_energy (int): Reference energy in eV
        ref_flux (float): Reference flux value
        ref_current (float): Reference current in mA

        Returns:
        float: Estimated flux
        """
        # Measured table at different beam parameters
        beam_sizes = np.array([300, 300, 300, 300, 200, 100, 50, 20, 9])
        pinhole_sizes = np.array([200, 100, 50, 20, 200, 200, 200, 200, 100])
        fluxes = np.array(
            [2e12, 5e11, 1.25e11, 2e10, 4.4e12, 9.9e12, 9.9e12, 9.9e12, 8.7e12]
        )
        energies = np.full(beam_sizes.shape, 12000)

        # Measured energy-flux dependence
        energy_flux = np.array(
            [
                [6000, 1.70e12],
                [7000, 3.70e12],
                [8000, 4.90e12],
                [10000, 3.90e12],
                [12000, 4.37e12],
                [14000, 4.20e12],
                [16000, 3.80e12],
                [18000, 3.00e12],
                [20000, 2.20e12],
                [22000, 1.60e12],
                [24000, 9.10e11],
                [26000, 5.80e11],
            ]
        )

        X = np.column_stack((beam_sizes, pinhole_sizes, energies))
        y = fluxes
        rbf_interpolator = Rbf(X[:, 0], X[:, 1], X[:, 2], y, function="linear")
        energy_flux_interpolator = interp1d(
            energy_flux[:, 0],
            energy_flux[:, 1],
            kind="linear",
            fill_value="extrapolate",
        )
        estimated_ref_flux = rbf_interpolator(
            ref_beam_size, ref_pinhole_size, ref_energy
        )

        if np.isnan(estimated_ref_flux):
            raise RuntimaWarning(
                "Reference flux cannot be estimated using the interpolator. Check reference points."
            )

        # Scale the flux based on the reference flux
        scale_factor = ref_flux / estimated_ref_flux

        # Estimate the flux at the desired beam, pinhole size, and energy
        estimated_flux = rbf_interpolator(
            beam_size, pinhole_size, 12000
        )  # Always interpolate at 12000 eV

        if np.isnan(estimated_flux):
            raise RuntimaWarning(
                "Desired flux cannot be estimated using the interpolator. Check input points."
            )

        # Scale the estimated flux based on the energy-dependent flux variation
        energy_scale_factor = energy_flux_interpolator(
            energy
        ) / energy_flux_interpolator(12000)  # Scaling relative to 12000 eV

        # Scale the flux based on the current
        current_scale_factor = current / ref_current

        return (
            estimated_flux * scale_factor * energy_scale_factor * current_scale_factor
        )



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Example yml configuration:

.. code-block:: yaml

 class: BlissMotor.BlissMotor
 configuration:
   actuator_name: dtox
   username: Detector Distance
"""

import enum

from bliss.config import static

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@enum.unique
class BlissMotorStates(enum.Enum):
    """
    MOVING = "Axis is moving"
    READY = "Axis is ready to be moved (not moving ?)"
    FAULT = "Error from controller"
    LIMPOS = "Hardware high limit active"
    LIMNEG = "Hardware low limit active"
    HOME = "Home signal active"
    OFF = "Axis power is off"
    DISABLED = "Axis cannot move"
    """

    MOVING = 0
    READY = 1
    FAULT = 2
    LIMPOS = 3
    LIMNEG = 4
    HOME = 5
    OFF = 6
    DISABLED = 7
    UNKNOWN = 8


class BlissMotor(AbstractMotor):
    """Bliss Motor implementation"""

    SPECIFIC_STATES = BlissMotorStates
    SPECIFIC_TO_HWR_STATE = {
        "MOVING": HardwareObjectState.BUSY,
        "READY": HardwareObjectState.READY,
        "FAULT": HardwareObjectState.FAULT,
        "LIMPOS": HardwareObjectState.READY,
        "LIMNEG": HardwareObjectState.READY,
        "HOME": HardwareObjectState.READY,
        "OFF": HardwareObjectState.OFF,
        "DISABLED": HardwareObjectState.OFF,
        "UNKNOWN": HardwareObjectState.UNKNOWN,
    }

    def __init__(self, name):
        super().__init__(name)
        self.motor_obj = None

    def init(self):
        """Initialise the motor"""
        super().init()
        cfg = static.get_config()
        self.motor_obj = cfg.get(self.actuator_name)

        # init state to match motor's one
        self.update_state(self.get_state())

        self.connect(self.motor_obj, "position", self.update_value)
        self.connect(self.motor_obj, "state", self._update_state)
        self.connect(self.motor_obj, "move_done", self._update_state)

    def _state2enum(self, state):
        """Translate the state to HardwareObjectState and BlissMotorStates
        Args:
           state (string): state
        Returns:
           (tuple): (HardwareObjectState, BlissMotorStates)
        """
        try:
            _specific_state = BlissMotorStates[state]
        except (TypeError, KeyError):
            _specific_state = BlissMotorStates.UNKNOWN

        _state = self.SPECIFIC_TO_HWR_STATE.get(state, HardwareObjectState.UNKNOWN)
        return _state, _specific_state

    def get_state(self):
        """Get the motor state.
        Returns:
            (enum HardwareObjectState): Motor state.
        """
        state = HardwareObjectState.UNKNOWN
        for stat in self.motor_obj.state.current_states_names:
            try:
                return HardwareObjectState[stat]
            except KeyError:
                if stat == "DISABLED":
                    # we need to treat DISABLED before any other auxiliary state
                    return HardwareObjectState.OFF
                if stat == "MOVING":
                    # MOVING has higher priority than other auxiliary states
                    return HardwareObjectState.BUSY
                # finally the state will corresponf to the last in the list
                # of the auxiliary states.
                state = self._state2enum(stat)[0]
        return state

    def get_specific_state(self):
        """Get the motor state.
        Returns:
            (list): Motor states as list of BlissMotorStates enum
        """
        state = self.motor_obj.state.current_states_names
        state_list = []
        for _state in state:
            state_list.append(self._state2enum(_state)[1])
        return state_list

    def _update_state(self, state=None):
        """Check if the state has changed. Emits signal stateChanged.
        Args:
            state (enum AxisState): state from a BLISS motor
        """
        if isinstance(state, bool):
            # It seems like the current version of BLISS gives us a boolean
            # at first and last event, True for ready and False for moving
            _state = HardwareObjectState.READY if state else HardwareObjectState.BUSY
        else:
            _state = self.get_state()
        # actualise the  self._specific_state every time
        self._specific_state = self.get_specific_state()
        # this will emit stateChanged if _state different from the previous one
        self.update_state(_state)

    def get_value(self):
        """Read the motor position.
        Returns:
            float: Motor position.
        """
        return self.motor_obj.position

    def get_limits(self):
        """Returns motor low and high limits.
        Returns:
            (tuple): two floats tuple (low limit, high limit).
        """
        # no limit = None, but None is a problematic value
        # for some GUI components (like MotorSpinBox), so
        # instead we return very large value.

        _low, _high = self.motor_obj.limits
        _low = _low if _low else -1e6
        _high = _high if _high else 1e6
        self._nominal_limits = (_low, _high)
        return self._nominal_limits

    def get_velocity(self):
        """Read motor velocity.
        Returns:
            (float): velocity [unit/s]
        """
        self._velocity = self.motor_obj.velocity
        return self._velocity

    def _set_value(self, value):
        """Move motor to absolute value.
        Args:
            value (float): target value
        """
        self.motor_obj.move(value, wait=False)

    def abort(self):
        """Stop the motor movement"""
        self.motor_obj.stop(wait=False)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

""" """

from mxcubecore.BaseHardwareObjects import HardwareObject

__author__ = "Ivars Karpics"
__credits__ = ["MXCuBE collaboration"]
__version__ = "2.2."


class BeamlineTools(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self.tools_list = []

    def init(self):
        for tool in self["tools"]:
            if tool.get_property("separator"):
                self.tools_list.append("separator")
            else:
                tool_dict = {
                    "hwobj": self.get_object_by_role(tool.get_property("hwobj")),
                    "display": tool.get_property("display"),
                    "method": tool.get_property("method"),
                }
                if tool.get_property("icon"):
                    tool_dict["icon"] = tool.get_property("icon")
                if tool.get_property("confirmation"):
                    tool_dict["confirmation"] = tool.get_property("confirmation")
                if tool.get_property("expertMode"):
                    tool_dict["expertMode"] = tool.get_property("expertMode")
                self.tools_list.append(tool_dict)

    def get_tools_list(self):
        return self.tools_list



# encoding: utf-8

#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
Utility class for importing queue entries
"""

import logging
import os
import sys
from importlib import import_module
from pathlib import Path


class ImportHelper:
    """
    Utility class for importing queue entries
    """

    MODULES = {}

    @staticmethod
    def import_queue_entry_modules(name, site_name=""):
        """
        Import all queue entries, using the convention that the class name
        is the camel cased modulename with a QueueEntry suffix.

        Args:
            name (str): Module name to attach queue entries to
            site_name (str): Site name, for looking up site specific
                             queue_entries

        Returns:
            dict: ImportHelper.MODULES (<str,class>)
        """

        package = f"mxcubecore.HardwareObjects.{site_name}.queue_entry"
        site_path = f"HardwareObjects/{site_name}/queue_entry"
        path = os.path.dirname(__file__).replace("queue_entry", site_path)

        package = package if site_name else __package__
        path = path if site_name else os.path.dirname(__file__)

        for f in Path(path).glob("*.py"):
            # Skip BaseQueueEntry, it is explicitly imported below as
            # the module contains several essential classes and helper
            # functions Skip xrf_spectrum to preserve casing (XRF) so that
            # we are backwards compatible (for the time being)
            if f.stem in [
                "import_helper",
                "base_queue_entry",
                "xrf_spectrum",
                "__init__",
            ]:
                continue
            m = import_module(f"{package}.{f.stem}")
            cls_name = f.stem.title().replace("_", "") + "QueueEntry"

            cls = getattr(m, cls_name, None)

            if cls:
                ImportHelper.MODULES[cls_name] = cls
                if not hasattr(sys.modules[name], cls_name):
                    setattr(sys.modules[name], cls_name, cls)
                    logging.getLogger("HWR").info(
                        f"Imported queue entry: {cls_name} from {f}"
                    )
                else:
                    logging.getLogger("HWR").warning(
                        f"Queue entry with name: {cls_name} already exists {f}"
                    )
            else:
                logging.getLogger("HWR").warning(
                    f"Could not find queue entry: {cls_name} in {f}"
                )

        return ImportHelper.MODULES



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import random
import time
import warnings

from gevent.event import AsyncResult

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.GenericDiffractometer import (
    GenericDiffractometer,
    PhaseEnum,
)


class DiffractometerMockup(GenericDiffractometer):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        GenericDiffractometer.__init__(self, name)

        # child object slots
        self.backlight = None
        self.backlightswitch = None
        self.beamstop_distance = None
        self.focus = None
        self.frontlight = None
        self.frontlightswitch = None
        self.kappa = None
        self.kappa_phi = None
        self.phi = None
        self.phiy = None
        self.phiz = None
        self.sampx = None
        self.sampy = None

    def init(self):
        """
        Descript. :
        """
        # self.image_width = 100
        # self.image_height = 100

        GenericDiffractometer.init(self)
        self.x_calib = 0.000444
        self.y_calib = 0.000446
        self.last_centred_position = [318, 238]

        self.pixels_per_mm_x = 1.0 / self.x_calib
        self.pixels_per_mm_y = 1.0 / self.y_calib
        self.beam_position = [318, 238]

        self.current_phase = GenericDiffractometer.PHASE_CENTRING

        self.cancel_centring_methods = {}
        self.current_motor_positions = {
            "phiy": 1.0,
            "sampx": 0.0,
            "sampy": -1.0,
            "zoom": 8.53,
            "focus": -0.42,
            "phiz": 1.1,
            "phi": 311.1,
            "kappa": 11,
            "kappa_phi": 22.0,
        }
        self.move_motors(self._get_random_centring_position())

        self.current_state_dict = {}
        self.centring_status = {"valid": False}
        self.centring_time = 0

        # self.image_width = 400
        # self.image_height = 400

        self.mount_mode = self.get_property("sample_mount_mode")
        if self.mount_mode is None:
            self.mount_mode = "manual"

        self.equipment_ready()

        self.connect(self.motor_hwobj_dict["phi"], "valueChanged", self.phi_motor_moved)
        self.connect(
            self.motor_hwobj_dict["phiy"], "valueChanged", self.phiy_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiz"], "valueChanged", self.phiz_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa"], "valueChanged", self.kappa_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa_phi"],
            "valueChanged",
            self.kappa_phi_motor_moved,
        )
        self.connect(
            self.motor_hwobj_dict["sampx"], "valueChanged", self.sampx_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["sampy"], "valueChanged", self.sampy_motor_moved
        )

    def execute_server_task(self, method, timeout=30, *args):
        return

    def in_plate_mode(self):
        return self.mount_mode == "plate"

    def use_sample_changer(self):
        return self.mount_mode == "sample_changer"

    def is_reversing_rotation(self):
        return True

    def get_grid_direction(self):
        """
        Descript. :
        """
        return self.grid_direction

    def manual_centring(self):
        """
        Descript. :
        """
        for click in range(3):
            self.user_clicked_event = AsyncResult()
            self.waiting_for_click = True
            x, y = self.user_clicked_event.get()
            if click < 2:
                self.motor_hwobj_dict["phi"].set_value_relative(90)
        self.last_centred_position[0] = x
        self.last_centred_position[1] = y
        centred_pos_dir = self._get_random_centring_position()
        return centred_pos_dir

    def automatic_centring(self):
        """Automatic centring procedure"""
        centred_pos_dir = self._get_random_centring_position()
        self.emit("newAutomaticCentringPoint", centred_pos_dir)
        return centred_pos_dir

    def _get_random_centring_position(self):
        """Get random centring result for current positions"""

        # Names of motors to vary during centring
        vary_actuator_names = ("sampx", "sampy", "phiy")

        # Range of random variation
        var_range = 0.08

        # absolute value limit for varied motors
        var_limit = 2.0

        result = self.current_motor_positions.copy()
        for tag in vary_actuator_names:
            val = result.get(tag)
            if val is not None:
                random_num = random.random()
                var = (random_num - 0.5) * var_range
                val += var
                if abs(val) > var_limit:
                    val *= 1 - var_range / var_limit
                result[tag] = val
        return result

    def is_ready(self) -> bool:
        """
        Descript. :
        """
        return True

    def is_valid(self):
        """
        Descript. :
        """
        return True

    def invalidate_centring(self):
        """
        Descript. :
        """
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            # self.emitProgressMessage("")
            self.emit("centringInvalid", ())

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """
        Descript. :
        """
        centred_pos_dir = self._get_random_centring_position()
        return centred_pos_dir

    def get_calibration_data(self, offset):
        """
        Descript. :
        """
        # return (1.0 / self.x_calib, 1.0 / self.y_calib)
        return (1.0 / self.x_calib, 1.0 / self.y_calib)

    def refresh_omega_reference_position(self):
        """
        Descript. :
        """
        return

    # def get_omega_axis_position(self):
    #     """
    #     Descript. :
    #     """
    #     return self.current_positions_dict.get("phi")

    def beam_position_changed(self, value):
        """
        Descript. :
        """
        self.beam_position = value

    def get_current_centring_method(self):
        """
        Descript. :
        """
        return self.current_centring_method

    def motor_positions_to_screen(self, centred_positions_dict):
        """
        Descript. :
        """
        return self.last_centred_position[0], self.last_centred_position[1]

    def moveToCentredPosition(self, centred_position, wait=False):
        """
        Descript. :
        """
        try:
            return self.move_to_centred_position(centred_position)
        except Exception:
            logging.exception("Could not move to centred position")

    def phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["phi"] = pos
        self.emit("phiMotorMoved", pos)

    def phiy_motor_moved(self, pos):
        self.current_motor_positions["phiy"] = pos

    def phiz_motor_moved(self, pos):
        self.current_motor_positions["phiz"] = pos

    def sampx_motor_moved(self, pos):
        self.current_motor_positions["sampx"] = pos

    def sampy_motor_moved(self, pos):
        self.current_motor_positions["sampy"] = pos

    def kappa_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaMotorMoved", pos)

    def kappa_phi_motor_moved(self, pos):
        """
        Descript. :
        """
        self.current_motor_positions["kappa_phi"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaPhiMotorMoved", pos)

    def refresh_video(self):
        """
        Descript. :
        """
        self.emit("minidiffStateChanged", "testState")
        if HWR.beamline.beam:
            HWR.beamline.beam.beam_pos_hor_changed(300)
            HWR.beamline.beam.beam_pos_ver_changed(200)

    def start_auto_focus(self):
        """
        Descript. :
        """
        return

    def move_to_beam(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """

        print(
            (
                "moving to beam position: %d %d"
                % (self.beam_position[0], self.beam_position[1])
            )
        )

    def move_to_coord(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """
        warnings.warn(
            "Deprecated method, call move_to_beam instead", DeprecationWarning
        )
        return self.move_to_beam(x, y, omega)

    def start_move_to_beam(self, coord_x=None, coord_y=None, omega=None):
        """
        Descript. :
        """
        self.last_centred_position[0] = coord_x
        self.last_centred_position[1] = coord_y
        self.centring_time = time.time()
        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centring_status = {
            "valid": True,
            "startTime": curr_time,
            "endTime": curr_time,
        }
        motors = self.get_positions()
        # motors["beam_x"] = 0.1
        # motors["beam_y"] = 0.1
        self.last_centred_position[0] = coord_x
        self.last_centred_position[1] = coord_y
        self.centring_status["motors"] = motors
        self.centring_status["valid"] = True
        self.centring_status["angleLimit"] = False
        self.emit_progress_message("")
        self.accept_centring()
        self.current_centring_method = None
        self.current_centring_procedure = None

    def re_emit_values(self):
        self.emit("zoomMotorPredefinedPositionChanged", None, None)
        omega_ref = [0, 238]
        self.emit("omegaReferenceChanged", omega_ref)

    def move_kappa_and_phi(self, kappa, kappa_phi):
        return

    def get_osc_max_speed(self):
        return 66

    def get_osc_limits(self):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        if self.in_plate_mode:
            return (170, 190)
        else:
            return (-360, 360)

    def get_osc_dynamic_limits(self):
        """Returns dynamic limits of oscillation axis"""
        return (0, 20)

    def get_scan_dynamic_limits(self, speed=None):
        return (-360, 360)

    def move_omega_relative(self, relative_angle):
        self.motor_hwobj_dict["phi"].set_value_relative(relative_angle, 5)

    def set_phase(self, phase, timeout=None):
        self.current_phase = str(phase)
        self.emit("minidiffPhaseChanged", (self.current_phase,))

    def get_point_from_line(self, point_one, point_two, index, images_num):
        return point_one.as_dict()

    def abort(self) -> None:
        return None

    def status(self) -> str:
        return "READY"

    def my_fancy_function(
        self, speed: float, num_images: int, exp_time: float, phase: PhaseEnum
    ) -> bool:
        return True

    def my_other_funny_function(self) -> None:
        pass

    def ssx_chip_scan(self, parameters):
        return

    def move_chip_to(self, x: int, y: int) -> None:
        print("moving chip to")
        return



import logging
import os
import time

import gevent
import gevent.event
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractEnergyScan import AbstractEnergyScan
from mxcubecore.TaskUtils import cleanup

scan_test_data = [
    (10841.0, 20.0),
    (10842.0, 20.0),
    (10843.0, 20.0),
    (10844.0, 20.0),
    (10845.0, 20.0),
    (10846.0, 20.0),
    (10847.0, 20.0),
    (10848.0, 20.0),
    (10849.0, 20.0),
    (10850.0, 20.0),
    (10851.0, 20.0),
    (10852.0, 20.0),
    (10853.0, 20.0),
    (10854.0, 20.0),
    (10855.0, 20.0),
    (10856.0, 20.0),
    (10857.0, 20.0),
    (10858.0, 20.00),
    (10859.0, 20.0),
    (10860.0, 20.0),
    (10861.0, 20.1),
    (10862.0, 21.4),
    (10863.0, 30.4),
    (10864.9, 80.7),
    (10865.9, 299.0),
    (10866.7, 820.8),
    (10867.5, 2009.2),
    (10868.2, 4305.5),
    (10869.0, 8070.2),
    (10869.8, 13246.7),
    (10870.6, 19124.1),
    (10871.4, 24430.5),
    (10872.2, 27843.1),
    (10873.0, 28654.8),
    (10873.8, 27092.5),
    (10874.6, 24138.5),
    (10875.4, 20957.3),
    (10876.0, 18373.6),
    (10877.0, 16373.8),
    (10878.0, 15474.8),
    (10879.0, 15163.9),
    (10880.0, 15080.5),
    (10881.0, 15063.0),
    (10882.0, 15060.3),
    (10883.0, 15059.8),
    (10884.0, 15059.7),
    (10885.0, 15059.0),
    (10886.0, 15059.0),
    (10887.0, 15059.0),
    (10888.0, 15059.0),
    (10889.0, 15059.0),
    (10890.0, 15059.0),
    (10891.0, 15059.0),
    (10892.0, 15059.0),
    (10893.0, 15059.0),
    (10894.0, 15059.0),
    (10895.0, 15059.0),
    (10896.0, 15059.0),
    (10897.0, 15059.0),
    (10898.0, 15059.0),
    (10899.0, 15059.0),
    (10900.0, 15059.0),
    (10901.0, 15059.0),
    (10902.0, 15059.0),
    (10903.0, 15059.0),
    (10904.0, 15059.0),
    (10905.0, 15059.0),
    (10906.0, 15059.0),
    (10907.0, 15059.0),
    (10908.0, 15059.0),
    (10909.0, 15059.0),
    (10910.0, 15059.0),
]

chooch_graph_data = (
    (12628, 0, -5),
    (12629, 0, -5),
    (12630, 0, -5),
    (12631, 0, -5),
    (12632, 0, -5),
    (12633, 0, -5),
    (12634, 0, -5),
    (12635, 0, -5),
    (12636, 0, -5),
    (12637, 0, -5),
    (12638, 0, -5),
    (12639, 0, -5),
    (12640, 0, -5),
    (12641, 0, -6),
    (12642, 0, -6),
    (12643, 0, -6),
    (12644, 0, -6),
    (12645, 0, -6),
    (12646, 0, -6),
    (12647, 0, -6),
    (12648, 0, -6),
    (12649, 0, -7),
    (12650, 0, -7),
    (12651, 0, -7),
    (12652, 0, -8),
    (12653, 0, -8),
    (12654, 0, -8),
    (12655, 1, -9),
    (12655, 2, -9),
    (12656, 3, -10),
    (12657, 4, -9),
    (12658, 5, -9),
    (12659, 6, -8),
    (12659, 6, -7),
    (12660, 6, -6),
    (12661, 5, -5),
    (12662, 4, -5),
    (12663, 4, -5),
    (12664, 3, -5),
    (12665, 3, -5),
    (12666, 3, -5),
    (12667, 3, -5),
    (12668, 3, -5),
    (12669, 3, -5),
    (12670, 3, -5),
    (12671, 3, -5),
    (12672, 3, -5),
    (12673, 3, -5),
    (12674, 3, -5),
    (12675, 3, -5),
    (12676, 3, -5),
    (12677, 3, -5),
    (12678, 3, -5),
    (12679, 3, -5),
    (12680, 3, -5),
    (12681, 3, -5),
    (12682, 3, -5),
    (12683, 3, -5),
    (12684, 3, -4),
    (12685, 3, -4),
    (12686, 3, -4),
    (12687, 3, -4),
    (12688, 3, -4),
    (12689, 3, -4),
    (12690, 3, -4),
    (12691, 3, -4),
    (12692, 3, -4),
    (12693, 3, -4),
    (12694, 3, -4),
    (12695, 3, -4),
    (12696, 3, -4),
)


class EnergyScanMockup(AbstractEnergyScan):
    def __init__(self, name):
        AbstractEnergyScan.__init__(self, name)

    def init(self):
        self.ready_event = gevent.event.Event()
        self.energy_scan_parameters = {}
        self.result_value_emitter = None
        self.scan_data = []
        self.thEdgeThreshold = 5
        self.energy2WavelengthConstant = 12.3980
        self.defaultWavelength = 0.976

    def emit_result_values(self):
        for value_tuple in scan_test_data:
            x = value_tuple[0]
            y = value_tuple[1]
            if not (x == 0 and y == 0):
                # if x is in keV, transform into eV otherwise let it like it is
                # if point larger than previous point (for chooch)
                if len(self.scan_data) > 0:
                    if x > self.scan_data[-1][0]:
                        self.scan_data.append([(x < 1000 and x * 1000.0 or x), y])
                else:
                    self.scan_data.append([(x < 1000 and x * 1000.0 or x), y])
                self.emit("scanNewPoint", (x < 1000 and x * 1000.0 or x), y)
            time.sleep(0.05)
        self.scanCommandFinished()

    def execute_energy_scan(self, energy_scan_parameters):
        self.energy_scan_parameters["exposureTime"] = 0.01
        print(self.cpos)
        if HWR.beamline.transmission is not None:
            self.energy_scan_parameters["transmissionFactor"] = (
                HWR.beamline.transmission.get_value()
            )
        else:
            self.energy_scan_parameters["transmissionFactor"] = None
        size_hor = None
        size_ver = None
        if HWR.beamline.beam is not None:
            size_hor, size_ver = HWR.beamline.beam.get_beam_size()
            size_hor = size_hor * 1000
            size_ver = size_ver * 1000
        self.energy_scan_parameters["beamSizeHorizontal"] = size_hor
        self.energy_scan_parameters["beamSizeVertical"] = size_ver
        self.energy_scan_parameters["startEnergy"] = 0
        self.energy_scan_parameters["endEnergy"] = 0
        self.energy_scan_parameters["fluorescenceDetector"] = "Mockup detector"
        self.scan_data = []
        self.result_value_emitter = gevent.spawn(self.emit_result_values)

    def do_chooch(self, elt, edge, scan_directory, archive_directory, prefix):
        """
        Descript. :
        """
        symbol = "_".join((elt, edge))
        scan_file_prefix = os.path.join(scan_directory, prefix)
        archive_file_prefix = os.path.join(archive_directory, prefix)

        if os.path.exists(scan_file_prefix + ".raw"):
            i = 1
            while os.path.exists(scan_file_prefix + "%d.raw" % i):
                i = i + 1
            scan_file_prefix += "_%d" % i
            archive_file_prefix += "_%d" % i

        scan_file_raw_filename = os.path.extsep.join((scan_file_prefix, "raw"))
        archive_file_raw_filename = os.path.extsep.join((archive_file_prefix, "raw"))
        scan_file_efs_filename = os.path.extsep.join((scan_file_prefix, "efs"))
        archive_file_efs_filename = os.path.extsep.join((archive_file_prefix, "efs"))
        scan_file_png_filename = os.path.extsep.join((scan_file_prefix, "png"))
        archive_file_png_filename = os.path.extsep.join((archive_file_prefix, "png"))

        try:
            if not os.path.exists(scan_directory):
                os.makedirs(scan_directory)
            if not os.path.exists(archive_directory):
                os.makedirs(archive_directory)
        except Exception:
            logging.getLogger("HWR").exception(
                "EnergyScan: could not create energy scan result directory."
            )
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return

        try:
            scan_file_raw = open(scan_file_raw_filename, "w")
            archive_file_raw = open(archive_file_raw_filename, "w")
        except Exception:
            logging.getLogger("HWR").exception(
                "EnergyScan: could not create energy scan result raw file"
            )
            self.store_energy_scan()
            self.emit("energyScanFailed", ())
            return
        else:
            scanData = []
            for i in range(len(self.scan_data)):
                x = float(self.scan_data[i][0])
                x = x < 1000 and x * 1000.0 or x
                y = float(self.scan_data[i][1])
                scanData.append((x, y))
                scan_file_raw.write("%f,%f\r\n" % (x, y))
                archive_file_raw.write("%f,%f\r\n" % (x, y))
            scan_file_raw.close()
            archive_file_raw.close()
            self.energy_scan_parameters["scanFileFullPath"] = str(
                scan_file_raw_filename
            )

        pk = 7.519
        ip = 7.516
        rm = 7.54
        fpPeak = -12.6
        fppPeak = 20.7
        fpInfl = -21.1
        fppInfl = 11.9
        comm = "Mockup results"
        self.energy_scan_parameters["edgeEnergy"] = 0.1
        self.thEdge = self.energy_scan_parameters["edgeEnergy"]
        logging.getLogger("HWR").info(
            "th. Edge %s ; chooch results are pk=%f, ip=%f, rm=%f"
            % (self.thEdge, pk, ip, rm)
        )

        self.energy_scan_parameters["peakEnergy"] = pk
        self.energy_scan_parameters["inflectionEnergy"] = ip
        self.energy_scan_parameters["remoteEnergy"] = rm
        self.energy_scan_parameters["peakFPrime"] = fpPeak
        self.energy_scan_parameters["peakFDoublePrime"] = fppPeak
        self.energy_scan_parameters["inflectionFPrime"] = fpInfl
        self.energy_scan_parameters["inflectionFDoublePrime"] = fppInfl
        self.energy_scan_parameters["comments"] = comm
        self.energy_scan_parameters["choochFileFullPath"] = scan_file_efs_filename
        self.energy_scan_parameters["filename"] = archive_file_raw_filename
        self.energy_scan_parameters["workingDirectory"] = archive_directory

        chooch_graph_x, chooch_graph_y1, chooch_graph_y2 = zip(*chooch_graph_data)
        chooch_graph_x = list(chooch_graph_x)
        for i in range(len(chooch_graph_x)):
            chooch_graph_x[i] = chooch_graph_x[i] / 1000.0

        title = "%s  %s  %s\n%.4f  %.2f  %.2f\n%.4f  %.2f  %.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fpPeak,
            fppPeak,
            ip,
            fpInfl,
            fppInfl,
        )
        fig = Figure(figsize=(15, 11))
        ax = fig.add_subplot(211)
        ax.set_title("%s" % title)
        ax.grid(True)
        ax.plot(*(zip(*self.scan_data)), **{"color": "black"})
        ax.set_xlabel("Energy")
        ax.set_ylabel("MCA counts")
        ax2 = fig.add_subplot(212)
        ax2.grid(True)
        ax2.set_xlabel("Energy")
        ax2.set_ylabel("")
        handles = []
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y1, color="blue"))
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y2, color="red"))
        canvas = FigureCanvasAgg(fig)

        self.energy_scan_parameters["jpegChoochFileFullPath"] = str(
            archive_file_png_filename
        )
        try:
            logging.getLogger("HWR").info(
                "Rendering energy scan and Chooch " + "graphs to PNG file : %s",
                scan_file_png_filename,
            )
            canvas.print_figure(scan_file_png_filename, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not print figure")
        try:
            logging.getLogger("HWR").info(
                "Saving energy scan to archive " + "directory for ISPyB : %s",
                archive_file_png_filename,
            )
            canvas.print_figure(archive_file_png_filename, dpi=80)
        except Exception:
            logging.getLogger("HWR").exception("could not save figure")

        self.store_energy_scan()

        logging.getLogger("HWR").info("<chooch> returning")
        self.emit(
            "choochFinished",
            (
                pk,
                fppPeak,
                fpPeak,
                ip,
                fppInfl,
                fpInfl,
                rm,
                chooch_graph_x,
                chooch_graph_y1,
                chooch_graph_y2,
                title,
            ),
        )
        return (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        )

    def get_elements(self):
        elements = []

        try:
            for el in self.config.elements["element"]:
                elements.append({"symbol": el["symbol"], "energy": el["energy"]})
        except IndexError:
            pass
        return elements

    #
    # def getDefaultMadEnergies(self):
    #     energies = []
    #     try:
    #         for el in self["mad"]:
    #             energies.append([float(el.energy), el.directory])
    #     except IndexError:
    #         pass
    #     return energies

    def is_connected(self):
        return True

    def scanCommandStarted(self, *args):
        """
        Descript. :
        """
        print(self.energy_scan_parameters)
        title = "%s %s: %s %s" % (
            self.energy_scan_parameters["sessionId"],
            self.energy_scan_parameters["blSampleId"],
            self.energy_scan_parameters["element"],
            self.energy_scan_parameters["edgeEnergy"],
        )
        dic = {
            "xlabel": "energy",
            "ylabel": "counts",
            "scaletype": "normal",
            "title": title,
        }
        self.emit("scanStart", dic)
        self.emit("energyScanStarted", dic)
        self.energy_scan_parameters["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = True

    def scanCommandFinished(self, *args):
        """
        Descript. :
        """
        with cleanup(self.ready_event.set):
            self.energy_scan_parameters["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
            logging.getLogger("HWR").debug("Energy Scan: finished")
            self.scanning = False
            self.energy_scan_parameters["startEnergy"] = self.scan_data[-1][0] / 1000.0
            self.energy_scan_parameters["endEnergy"] = self.scan_data[-1][1] / 1000.0
            self.emit("energyScanFinished", self.energy_scan_parameters)

    def get_scan_data(self):
        """
        Descript. :
        """
        return self.scan_data

    def store_energy_scan(self):
        """
        Descript. :
        """
        if HWR.beamline.lims:
            db_status = HWR.beamline.lims.storeEnergyScan(self.energy_scan_parameters)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Example xml_ configuration:

.. code-block:: xml

 <object class="ApertureMockup">
   <username>aperture</username>
   <values>{"A5": (5, 0.11), A10": (10, 0.15), "A20": (20, 0.3), "A30": (30, 0.63), "A50": (50, 0.9), "A100": (100, 1)}</values>
   <position_list>["BEAM", "OFF", "PARK"]</position_list>
 </object>
"""

from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState
from mxcubecore.HardwareObjects.mockup.ActuatorMockup import ActuatorMockup

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ApertureMockup(AbstractNState, ActuatorMockup):
    """Mockup file for aperture as Nstate actuator"""

    def init(self):
        super().init()
        # check if we have values other that UNKNOWN (no values in config)
        if len(self.VALUES) == 1:
            self._initialise_values()
            self._initialise_inout()
        self._nominal_value = self.VALUES.A10

    def get_value(self):
        return self._nominal_value

    def _initialise_values(self):
        """Initialise the ValueEnum if not in the config"""
        values = {}
        predefined_postions = (5, 10, 20, 30, 50, 100)
        factors = (0.11, 0.15, 0.3, 0.63, 0.9, 1)
        for _pos, _fac in zip(predefined_postions, factors):
            values[f"A{_pos}"] = (_pos, _fac)
        self.VALUES = Enum(
            "ValueEnum",
            dict(values, **{item.name: item.value for item in self.VALUES}),
        )

    def _initialise_inout(self):
        """Add IN and OUT to the values Enum"""
        values_dict = {"IN": "BEAM", "OUT": "OFF"}
        values_dict.update({item.name: item.value for item in self.VALUES})
        self.VALUES = Enum("ValueEnum", values_dict)

    def get_factor(self, label):
        """Get the factor associated to a label.
        Args:
            (enum, str): label enum or name
        Returns:
            (float) or (tuple): Factor value
        """
        if isinstance(label, str):
            try:
                return self.VALUES[label].value[1]
            except (KeyError, ValueError, IndexError):
                return 1.0
        try:
            return label.value[1]
        except (ValueError, IndexError):
            return 1.0

    def get_size(self, label):
        """Get the aperture size associated to a label.
        Args:
            (enum, str): label enum or name
        Returns:
            (float): Factor value
        Raises:
            RuntimeError: Unknown aperture size.
        """
        if isinstance(label, str):
            try:
                return float(self.VALUES[label].value[0])
            except (KeyError, ValueError, IndexError) as err:
                raise RuntimeError("Unknown aperture size") from err
        try:
            return float(label.value[0])
        except (ValueError, IndexError) as err:
            if self.inout_obj:
                return None
            raise RuntimeError("Unknown aperture size") from err

    def get_diameter_size_list(self):
        """Get the list of values to be visible. Hide IN, OUT and UNKNOWN.
        Returns:
            (list): List of available aperture values (string).
        """
        values = []
        for value in self.VALUES:
            _nam = value.name

            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                values.append(_nam)

        return values



from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.ExporterMotor import ExporterMotor


class MicrodiffLight(ExporterMotor):
    def __init__(self, name):
        ExporterMotor.__init__(self, name)
        self._motor_pos_suffix = "Level"

    def init(self):
        ExporterMotor.init(self)
        try:
            _low, _high = self.get_property("limits").split(",")
            self._limits = (float(_low), float(_high))
        except (AttributeError, TypeError, ValueError):
            self._limits = (0, 10)
        self.chan_light_is_on = self.get_channel_object("chanLightIsOn")
        self.update_state(self.STATES.READY)

    def get_state(self):
        """Get the light state as a motor.
        Returns:
            (enum 'HardwareObjectState'): Light state.
        """
        return self._state

    def get_limits(self):
        return self._limits

    def light_is_out(self):
        return self.chan_light_is_on.get_value()

    def move_in(self):
        self.chan_light_is_on.set_value(True)

    def move_out(self):
        self.chan_light_is_on.set_value(False)

    def _set_value(self, value):
        """Move motor to absolute value.
        Args:
            value (float): target value
        """
        self.update_state(self.STATES.BUSY)
        self.motor_position_chan.set_value(value)
        self.update_state(self.STATES.READY)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Mock-up class to simulate the beamline flux, used for testing.
"""

from random import random

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractFlux import AbstractFlux

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class FluxMockup(AbstractFlux):
    """Class to simulate beamline flux"""

    def __init__(self, name):
        super().__init__(name)

        self.measured_flux_list = []
        self.measured_flux_dict = {}
        self.current_flux_dict = {}

    def init(self):
        super().init()
        self.current_flux_dict["flux"] = self.default_value
        gevent.spawn(self._update_flux)

    def _update_flux(self):
        while True:
            self.measure_flux()
            gevent.sleep(3)

    def get_value(self):
        """Get flux at current transmission in units of photons/s"""
        return self.current_flux_dict["flux"]

    def measure_flux(self) -> None:
        """
        measures intensity

        Emits:
           valueChanged (float): The new flux value
        """
        beam_size = HWR.beamline.beam.get_beam_size()
        transmission = HWR.beamline.transmission.get_value()
        flux = self.default_value * (1 + 0.001 * random()) * transmission / 100.0

        self.measured_flux_list = [
            {
                "size_x": beam_size[0],
                "size_y": beam_size[1],
                "transmission": transmission,
                "flux": flux,
            }
        ]

        self.measured_flux_dict = self.measured_flux_list[0]
        self.current_flux_dict = self.measured_flux_list[0]

        self.emit("valueChanged", self.current_flux_dict["flux"])

    @property
    def is_beam(self):
        """Check if there is beam
        Returns:
            (bool): True if beam present, False otherwise
        """
        return True

    def wait_for_beam(self, timeout=None):
        """Wait until beam present
        Args:
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait
                                              (default);
                             if timeout is None: wait forever.
        """
        with gevent.Timeout(timeout, RuntimeError("Timeout while waiting for beam")):
            gevent.sleep(timeout + 1)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


"""EMBLExporterClient"""

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command import Exporter

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLExporterClient(HardwareObject):
    """EMBLExport client"""

    def __init__(self, *args):
        HardwareObject.__init__(self, *args)
        self.exporter_client = None

    def init(self):
        for client in Exporter.exporter_clients:
            self.exporter_client = Exporter.exporter_clients[client]
            return

    def get_client_info(self):
        """
        Returns client info
        :return:
        """
        for exporter_client in Exporter.exporter_clients:
            return exporter_client

    def get_method_list(self):
        """
        Returns list with available methods
        :return: list
        """
        method_list = []
        for exporter_client in Exporter.exporter_clients:
            client = Exporter.exporter_clients[exporter_client]
            method_list.extend(client.get_method_list())
        return method_list

    def get_property_list(self):
        """
        Returns list with properties
        :return:
        """
        property_list = []
        for exporter_client in Exporter.exporter_clients:
            client = Exporter.exporter_clients[exporter_client]
            property_list.extend(client.get_propertyList())
        return property_list

    def read_property(self, property_name):
        """
        Returns property value
        :param property_name: str
        :return:
        """
        return self.exporter_client.readProperty(property_name)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent
from tine import query as tinequery

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLBeamFocusing(HardwareObject):
    """Hardware Object is used to evaluate and set beam focusing mode."""

    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.aperture_hwobj = None
        self.active_focus_mode = None
        self.size = [9999, 9999]
        self.focus_modes = None
        self.focus_motors_dict = None
        self.motors_groups = []

        self.cmd_set_calibration_name = None
        self.cmd_set_phase = None

    def init(self):
        """Reads available focusing modes from the config xml and
        attaches corresponding motors
        """

        self.cmd_set_calibration_name = self.get_command_object("cmdSetCalibrationName")
        self.focus_modes = []
        for focus_mode in self["focusModes"]:
            self.focus_modes.append(
                {
                    "modeName": focus_mode.modeName,
                    "lensCombination": eval(focus_mode.lensCombination),
                    "aperture": focus_mode.aperture,
                    "lensModes": eval(focus_mode.lensModes),
                    "size": eval(focus_mode.size),
                    "message": eval(focus_mode.message),
                    "diverg": eval(focus_mode.divergence),
                }
            )
        self.focus_motors_dict = {}

        focus_motors = eval(self.get_property("focusMotors", "[]"))

        for focus_motor in focus_motors:
            self.focus_motors_dict[focus_motor] = []

        self.motors_groups = [
            self.get_object_by_role("P14ExpTbl"),
            self.get_object_by_role("P14KB"),
            self.get_object_by_role("P14DetTrans"),
            self.get_object_by_role("P14BCU"),
            self.get_object_by_role("slitsMotors"),
        ]

        if len(self.motors_groups) > 0:
            for motors_group in self.motors_groups:
                self.connect(
                    motors_group,
                    "mGroupFocModeChanged",
                    self.motor_group_focus_mode_changed,
                )
                motors_group.re_emit_values()
        else:
            logging.getLogger("HWR").debug("BeamFocusing: No motors defined")
            self.active_focus_mode = self.focus_modes[0]["modeName"]
            self.size = self.focus_modes[0]["size"]
        self.re_emit_values()

        try:
            self.cmd_set_phase = eval(self.get_property("setPhaseCmd"))
        except Exception:
            pass

        self.aperture_hwobj = self.get_object_by_role("aperture")

    def get_focus_motors(self):
        """Returns a list with all focusing motors"""

        focus_motors = []
        if self.motors_groups is not None:
            for motors_group in self.motors_groups:
                motors_group_list = motors_group.get_motors_dict()
                for motor in motors_group_list:
                    focus_motors.append(motor)
        return focus_motors

    def motor_group_focus_mode_changed(self, value):
        """Called if motors group focusing is changed

        :param value: focusing mode
        :type value: str or None
        """
        motors_group_foc_mode = value
        for motor in motors_group_foc_mode:
            if motor in self.focus_motors_dict:
                self.focus_motors_dict[motor] = motors_group_foc_mode[motor]

        prev_mode = self.active_focus_mode
        self.active_focus_mode, self.size = self.get_active_focus_mode()

        if prev_mode != self.active_focus_mode:
            if self.active_focus_mode:
                logging.getLogger("GUI").info(
                    "Focusing: %s mode detected" % self.active_focus_mode
                )
            self.emit("focusingModeChanged", self.active_focus_mode, self.size)
            if self.active_focus_mode:
                self.cmd_set_calibration_name(self.active_focus_mode.lower())

    def get_focus_mode_names(self):
        """Returns defined focus modes names"""
        names = []
        for focus_mode in self.focus_modes:
            names.append(focus_mode["modeName"])
        return names

    def get_focus_mode_message(self, focus_mode_name):
        """Returns messages used when a new focusing mode is requisted.

        :param focus_mode_name: name of the mode
        :type focus_mode_name: str
        """
        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == focus_mode_name:
                return focus_mode["message"]

    def get_available_lens_modes(self, focus_mode_name=None):
        """Get available CRL lens combination for the given focusing mode

        :param focus_mode_name: requested focusing mode. If None passed then
                                current focusing mode is used
        :type focus_mode_name: str
        """
        lens_modes = ["Manual"]

        if focus_mode_name is None:
            focus_mode_name = self.active_focus_mode
        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == focus_mode_name:
                lens_modes = focus_mode["lensModes"]

        return lens_modes

    def get_lens_combination(self, focus_mode_name=None):
        """Returns available lens combination for the given focusing mode

        :param focus_mode_name: requested focusing mode. If None passed then
                                current focusing mode is used
        :type focus_mode_name: str
        """
        if focus_mode_name is None:
            focus_mode_name, beam_size = self.get_active_focus_mode()

        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == focus_mode_name:
                return focus_mode["lensCombination"]

    def get_focus_mode_aperture(self, focus_mode_name=None):
        """
        Returns aperture associated to the current beam focus mode
        :param focus_mode_name: diameter in micons
        :return:
        """
        if focus_mode_name is None:
            focus_mode_name, beam_size = self.get_active_focus_mode()

        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == focus_mode_name:
                return focus_mode["aperture"]

    def get_active_focus_mode(self):
        """Evaluates and returns active focusing mode"""
        if len(self.focus_motors_dict) > 0:
            active_focus_mode = None
            for focus_mode in self.focus_modes:
                self.size = focus_mode["size"]
                active_focus_mode = focus_mode["modeName"]
                for motor in self.focus_motors_dict:
                    if len(self.focus_motors_dict[motor]) == 0:
                        active_focus_mode = None
                        self.size = [9999, 9999]
                    elif active_focus_mode not in self.focus_motors_dict[motor]:
                        active_focus_mode = None
                        self.size = [9999, 9999]
                        break
                if active_focus_mode is not None:
                    break
            if active_focus_mode != self.active_focus_mode:
                self.active_focus_mode = active_focus_mode
        return self.active_focus_mode, self.size

    def get_focus_mode(self):
        """Returns active focusing mode"""
        if self.active_focus_mode:
            return self.active_focus_mode.lower()

    def set_motor_focus_mode(self, motor_name, focus_mode):
        """Sets focusing mode of a selected motor

        :param motor_name: motor name
        :type motor_name: str
        :param focus_mode: requested focusing mode
        :type focus_mode: str
        """
        if focus_mode is not None:
            for motor in self.motors_groups:
                motor.set_motor_focus_mode(motor_name, focus_mode)

    def set_focus_mode(self, focus_mode):
        """Sets focusing mode to all motors

        :param focus_mode: requested focusing mode
        :type focus_mode: str
        """
        gevent.spawn(self.focus_mode_task, focus_mode)
        logging.getLogger("HWR").info("Focusing: %s mode requested" % focus_mode)
        self.emit("focusingModeRequested", focus_mode)

    def focus_mode_task(self, focus_mode):
        """Gevent task to set focusing mode

        :param focus_mode: requested focusing mode
        :type focus_mode: str
        """
        if focus_mode and self.cmd_set_phase:
            # Waits for diffractometer to be ready
            self.aperture_hwobj.wait_ready()
            if focus_mode != "Imaging":
                logging.getLogger("GUI").warning(
                    "Focusing: Setting diffractometer to BeamLocation phase..."
                )

                tinequery(
                    self.cmd_set_phase["address"],
                    self.cmd_set_phase["property"],
                    self.cmd_set_phase["argument"],
                )

            logging.getLogger("GUI").warning("Focusing: Setting focusing motors...")
            if self.motors_groups:
                for motors_group in self.motors_groups:
                    motors_group.set_motor_group_focus_mode(focus_mode)
            logging.getLogger("GUI").info("Focusing: Focusing motors set")

            aperture_diameter = self.get_focus_mode_aperture(focus_mode)
            logging.getLogger("GUI").warning(
                "Focusing: Setting aperture to %d microns..." % aperture_diameter
            )

            self.aperture_hwobj.wait_ready()
            self.aperture_hwobj.set_diameter(aperture_diameter)
            logging.getLogger("GUI").info("Focusing: Aperture set")
        else:
            # No motors defined
            self.active_focus_mode = focus_mode

    def get_divergence_hor(self):
        """Returns horizontal beam divergence"""
        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == self.active_focus_mode:
                return focus_mode["diverg"][0]

    def get_divergence_ver(self):
        """Returns vertical beam divergence"""
        for focus_mode in self.focus_modes:
            if focus_mode["modeName"] == self.active_focus_mode:
                return focus_mode["diverg"][1]

    def re_emit_values(self):
        """Reemits available signals"""
        self.emit("focusingModeChanged", self.active_focus_mode, self.size)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name] EMBLMotorsGroup

[Description]
The MotorsGroup Hardware Object is used to maintain several motors in one
group. Motors group is a list of motors which are like a grouped instance
in tine server (a tuple). It allows to read several motor position,
statuses,... by one read.

[Channels]
- self.chanPositions
- self.chanStatus

[Commands]
- implemented as tine.set

[Emitted signals]
- mGroupPosChanged
- mGroupFocModeChanged
- mGroupStatusChanged

[Functions]
- setMotorPosition()
- setMotorFocMode()
- setMotorGroupFocMode()
- stopMotor()
- positionsChanged()
- statusChanged()

[Included Hardware Objects] -

Example Hardware Object XML file :
==================================
<object class="MotorsGroup">
    <username>P14BCU</username>                     - used to identify group
    <serverAddr>/P14/P14BCU</serverAddr>            - tine server address
    <groupAddr>/ShutterTrans</groupAddr>            - motors group address
    <positionAddr>Position</positionAddr>           - position address
    <statusAddr>Status</statusAddr>                 - status address
    <motors>                                        - motors list
        <motor>
          <motorName>ShutterTrans</motorName>       - name
          <motorAddr>ShutterTrans</motorAddr>       - address
          <setCmd>MOVE.start</setCmd>               - set cmd
          <stopCmd>MOVE.stop</stopCmd>              - stop cmd
          <index>0</index>                          - index in the group
          <velocity>None</velocity>                 - velocity
          <updateTolerance>0.005</updateTolerance>  - absolute update tolerance
          <evalTolerance>0.005</evalTolerance>      - absolute tolerance of
                                                      beam focus mode evaluation
          <statusModes>{'Move': 1, 'Ready': 0}</statusModes>
          <focusingModes>{'Collimated': 0.22, 'Horizontal': 0.22,
          'Vertical': 0.22, 'Double': 0.22}</focusingModes>
        </motor>
    </motors>
</object>
"""

import logging
import time

import gevent
import tine

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "Motor"


class EMBLMotorsGroup(HardwareObject):
    """
    EMBLMotorsGroup
    """

    def __init__(self, name):
        super().__init__(name)
        self.server_address = None
        self.group_address = None
        self.motors_list = None
        self.motors_group_position_dict = None
        self.motors_group_status_dict = None
        self.motors_group_foc_mode_dict = None

        self.chan_positions = None
        self.chan_status = None

    def init(self):
        self.motors_group_position_dict = {}
        self.motors_group_status_dict = {}
        self.motors_group_foc_mode_dict = {}

        self.server_address = self.serverAddr
        self.group_address = self.groupAddr
        self.motors_list = []

        for motor in self["motors"]:
            temp_dict = {}
            temp_dict["motorName"] = motor.motorName
            temp_dict["motorAddr"] = motor.motorAddr
            temp_dict["setCmd"] = motor.setCmd
            temp_dict["index"] = motor.index
            temp_dict["velocity"] = motor.velocity
            temp_dict["updateTolerance"] = motor.updateTolerance
            temp_dict["evalTolerance"] = motor.evalTolerance
            temp_dict["statusModes"] = eval(motor.statusModes)
            temp_dict["focusingModes"] = eval(motor.focusingModes)
            temp_dict["status"] = None
            temp_dict["position"] = -9999
            temp_dict["focMode"] = []
            self.motors_list.append(temp_dict)

        try:
            self.chan_positions = self.add_channel(
                {
                    "type": "tine",
                    "tinename": self.server_address + self.group_address,
                    "name": self.positionAddr,
                },
                self.positionAddr,
            )
            self.chan_positions.connect_signal("update", self.positions_changed)
            self.positions_changed(self.chan_positions.get_value())
        except Exception:
            msg = "EMBLMotorsGroup: unable to add channel %s/%s %s" % (
                self.server_address,
                self.group_address,
                self.positionAddr,
            )
            logging.getLogger("HWR").error(msg)

        try:
            self.chan_status = self.add_channel(
                {
                    "type": "tine",
                    "tinename": self.server_address + self.group_address,
                    "name": self.statusAddr,
                },
                self.statusAddr,
            )
            self.chan_status.connect_signal("update", self.status_changed)
            self.status_changed(self.chan_status.get_value())
        except Exception:
            msg = "EMBLMotorsGroup: unable to add channel %s/%s %s" % (
                self.server_address,
                self.group_address,
                self.statusAddr,
            )
            logging.getLogger("HWR").error(msg)

    def get_motors_dict(self):
        """Returns dict with motors"""
        return self.motors_list

    def set_motor_position(self, motor_name, new_position, timeout=None):
        """Sets motor value. Direct tine.set cmd is used"""
        for motor in self.motors_list:
            if motor["motorName"] == motor_name:
                if motor["velocity"] is not None:
                    tine.set(
                        self.server_address + "/" + motor["motorAddr"],
                        "Velocity",
                        motor["velocity"],
                    )
                motor["status"] = motor["statusModes"]["Move"]
                tine.set(
                    self.server_address + "/" + motor["motorAddr"],
                    motor["setCmd"],
                    new_position,
                )
                logging.getLogger("HWR").debug(
                    "EMBLMotorsGroup: send %s : %.4f"
                    % (motor["motorAddr"], new_position)
                )
                time.sleep(0.2)
                self.wait_motor_ready(motor_name, timeout=10)
                time.sleep(1)
                logging.getLogger("HWR").debug(
                    "EMBLMotorsGroup: motor %s ready" % motor["motorAddr"]
                )
                break

    def set_motor_focus_mode(self, motor_name, focus_mode):
        """Sets a focus mode for an individual motor"""
        for motor in self.motors_list:
            if motor["motorName"] == motor_name:
                if (
                    motor["setCmd"] is not None
                    and focus_mode in motor["focusingModes"].keys()
                ):
                    if motor["velocity"] is not None:
                        tine.set(
                            self.server_address + "/" + motor["motorAddr"],
                            "Velocity",
                            motor["velocity"],
                        )
                    tine.set(
                        self.server_address + "/" + motor["motorAddr"],
                        motor["setCmd"],
                        motor["focusingModes"][focus_mode],
                    )
                    time.sleep(1)
                break

    def set_motor_group_focus_mode(self, focus_mode):
        """Sets a focus mode for the motors group"""
        for motor in self.motors_list:
            if (
                motor["setCmd"] is not None
                and focus_mode in motor["focusingModes"].keys()
            ):
                if motor["velocity"] is not None:
                    tine.set(
                        self.server_address + "/" + motor["motorAddr"],
                        "Velocity",
                        motor["velocity"],
                    )
        time.sleep(0.5)

        for motor in self.motors_list:
            if (
                motor["setCmd"] is not None
                and focus_mode in motor["focusingModes"].keys()
            ):
                motor["status"] = motor["statusModes"]["Move"]
                tine.set(
                    self.server_address + "/" + motor["motorAddr"],
                    motor["setCmd"],
                    motor["focusingModes"][str(focus_mode)],
                )
                logging.getLogger("HWR").debug(
                    "EMBLMotorsGroup: send %s : %.4f"
                    % (motor["motorAddr"], motor["focusingModes"][str(focus_mode)])
                )
                if motor["motorName"] in ("In", "Out", "Top", "But"):
                    self.wait_motor_ready(motor["motorName"], timeout=10)
                    time.sleep(1.1)
                    logging.getLogger("HWR").debug(
                        "EMBLMotorsGroup: motor %s ready" % motor["motorAddr"]
                    )

    def stop_motor(self, motor_name):
        """Stops motor movement"""
        for motor in self.motors_list:
            if motor["motorName"] == motor_name:
                if motor["setCmd"] is not None:
                    tine.set(
                        self.server_address + self.group_address + "/" + motor_name,
                        motor["stopCmd"],
                    )
                break

    def positions_changed(self, positions):
        """Called if one or several motors values has been changed.
        Evaluates if value needs to be updates, if value is
        changed, then evaluates focusing mode. If necessary
        pysignals are emitted
        """
        do_emit = False
        # values_to_send = {}
        # foc_mode_to_send = {}
        for motor in self.motors_list:
            old_value = motor["position"]
            if isinstance(positions, (list, tuple)):
                new_value = positions[motor["index"]]
            else:
                new_value = positions
            if abs(old_value - new_value) > motor["updateTolerance"]:
                motor["position"] = new_value
                do_emit = True
            if do_emit:
                self.motors_group_position_dict[motor["motorName"]] = new_value
                motor["focMode"] = []
                for foc_mode in motor["focusingModes"]:
                    diff = abs(motor["focusingModes"][foc_mode] - new_value)
                    if diff < motor["evalTolerance"]:
                        motor["focMode"].append(foc_mode)
                self.motors_group_foc_mode_dict[motor["motorName"]] = motor["focMode"]
        if do_emit:
            self.emit("mGroupPosChanged", self.motors_group_position_dict)
            self.emit("mGroupFocModeChanged", self.motors_group_foc_mode_dict)

    def get_detected_foc_mode(self):
        """
        Returns focus mode
        :return: str
        """
        return self.detected_foc_mode

    def status_changed(self, status):
        """Called if motors status is changed. Pysignal with new
        status has been sent"""
        for motor in self.motors_list:
            old_status = motor["status"]
            if isinstance(status, (list, tuple)):
                new_status = status[motor["index"]]
            else:
                new_status = status
            if old_status != new_status:
                motor["status"] = new_status
                for status_mode in motor["statusModes"]:
                    if motor["statusModes"][status_mode] == new_status:
                        self.motors_group_status_dict[motor["motorName"]] = status_mode
        self.emit("mGroupStatusChanged", self.motors_group_status_dict)

    def wait_motor_ready(self, motor_name, timeout):
        """Waits motor ready"""
        self.status_changed(self.chan_status.get_value())
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self.is_motor_ready(motor_name):
                gevent.sleep(0.01)

    def is_motor_ready(self, motor_name):
        """Returns True if motors is ready"""
        is_ready = False
        for motor in self.motors_list:
            if motor["motorName"] == motor_name:
                is_ready = motor["status"] == motor["statusModes"]["Ready"]
                break
        return is_ready

    def re_emit_values(self):
        """
        Reemits all signals
        :return:
        """
        self.emit("mGroupPosChanged", self.motors_group_position_dict)
        self.emit("mGroupFocModeChanged", self.motors_group_foc_mode_dict)
        self.emit("mGroupStatusChanged", self.motors_group_status_dict)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import time
from datetime import datetime

import yaml

from mxcubecore.BaseHardwareObjects import HardwareObject

__author__ = "EMBL Hamburg"
__credits__ = ["MXCuBE collaboration"]
__version__ = "2.3."


class StateMachine(HardwareObject):
    """Finite State Machine (FSM) is a mathematical model of a closed or
    opened loop discrete-event systems with well defined state.
    It is wildly used to define functioning system and control their
    execution. In the case of MX beamlines and MXCuBE FSM represents
    different state where certain action from a user is requested.
    It is possible to describe a sequence of user actions as a discrete
    state that are logically connected.
    For example, each MX experiment requires a crystal to be mounted on a
    goniostat. If not crystal is mounted then it makes no sense to
    continue an experiment.
    The actual transition logic is implemented in the update_fsm_state()
    """

    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.state_list = []
        self.condition_list = None
        self.transition_list = None
        self.current_state = None
        self.previous_state = ""
        self.history_state_list = []

    def init(self):
        with open(self.get_property("structure_file"), "r") as stream:
            data_loaded = yaml.load(stream)

        self.state_list = data_loaded["states"]
        self.condition_list = data_loaded["conditions"]
        self.transition_list = data_loaded["transitions"]
        self.current_state = data_loaded["initial_state"]
        self.previous_state = data_loaded["initial_state"]

        for condition in self.condition_list:
            condition["value"] = False
            if "desc" not in condition:
                condition["desc"] = condition["name"].title().replace("_", " ")

        for transition in self.transition_list:
            if not self.get_state_by_name(transition["source"]):
                logging.getLogger("HWR").error(
                    "Transition %s " % str(transition)
                    + "has a none existing source state: %s" % transition["source"]
                )
            if not self.get_state_by_name(transition["dest"]):
                logging.getLogger("HWR").error(
                    "Transition %s " % str(transition)
                    + "has a none existing destination state: %s" % transition["dest"]
                )

            if "conditions_true" not in transition:
                transition["conditions_true"] = []
            if "conditions_false" not in transition:
                transition["conditions_false"] = []
            if "conditions_false_or" not in transition:
                transition["conditions_false_or"] = []

            for condition_name in transition["conditions_true"]:
                if not self.get_condition_by_name(condition_name):
                    logging.getLogger("HWR").error(
                        "Transition %s " % str(transition)
                        + "has a none existing condition: %s" % condition_name
                    )
            for condition_name in transition["conditions_false"]:
                if not self.get_condition_by_name(condition_name):
                    logging.getLogger("HWR").error(
                        "Transition %s " % str(transition)
                        + "has a none existing condition: %s" % condition_name
                    )
            for condition_name in transition["conditions_false_or"]:
                if not self.get_condition_by_name(condition_name):
                    logging.getLogger("HWR").error(
                        "Transition %s " % str(transition)
                        + "has a none existing condition: %s" % condition_name
                    )

        self.update_fsm_state()

        self.bl_setup_hwobj = self.get_object_by_role("beamline_setup")
        for hwobj_name in dir(self.bl_setup_hwobj):
            if hwobj_name.endswith("hwobj"):
                # logging.getLogger("HWR").debug(\
                #     "StateMachine: Attaching hwobj: %s " % hwobj_name)
                self.connect(
                    getattr(self.bl_setup_hwobj, hwobj_name),
                    "fsmConditionChanged",
                    self.condition_changed,
                )
                getattr(self.bl_setup_hwobj, hwobj_name).re_emit_values()

    def get_state_by_name(self, state_name):
        for state in self.state_list:
            if state["name"] == state_name:
                return state

    def get_condition_by_name(self, condition_name):
        for condition in self.condition_list:
            if condition["name"] == condition_name:
                return condition

    def condition_changed(self, condition_name, value):
        """Event when condition of a hardware object has been changed"""

        condition = self.get_condition_by_name(condition_name)
        if condition:
            # logging.getLogger("HWR").debug(\
            #  "StateMachine: condition '%s' changed to '%s'" \
            #   % (condition_name, value))

            if condition["value"] != value:
                condition["value"] = value
                self.emit("conditionChanged", self.condition_list)
                self.update_fsm_state()
        else:
            logging.getLogger("HWR").debug(
                "StateMachine: condition '%s' not in the condition list"
                % condition_name
            )

    def update_fsm_state(self):
        """Updates state machine
        We look at the current state and available transitions from it
        If all conditions of a transition is met then the transition is
        executed and signal is emitted.
        """
        for transition in self.transition_list:
            if transition["source"] == self.current_state:
                allow_transition = True
                for cond_name in transition["conditions_true"]:
                    cond = self.get_condition_by_name(cond_name)
                    if cond["value"] is False:
                        allow_transition = False
                for cond_name in transition["conditions_false"]:
                    cond = self.get_condition_by_name(cond_name)
                    if cond["value"] is True:
                        allow_transition = False
                for cond_name in transition["conditions_false_or"]:
                    cond = self.get_condition_by_name(cond_name)
                    if cond["value"] is False:
                        allow_transition = True

                if allow_transition:
                    self.current_state = transition["dest"]
                    break

        if self.previous_state != self.current_state:
            if self.history_state_list:
                self.history_state_list[-1]["end_time"] = time.strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
                self.history_state_list[-1]["total_time"] = str(
                    datetime.strptime(
                        self.history_state_list[-1]["end_time"], "%Y-%m-%d %H:%M:%S"
                    )
                    - datetime.strptime(
                        self.history_state_list[-1]["start_time"], "%Y-%m-%d %H:%M:%S"
                    )
                )

            history_state_item = {
                "current_state": self.current_state,
                "previous_state": self.previous_state,
                "start_time": time.strftime("%Y-%m-%d %H:%M:%S"),
                "end_time": "... ",
                "total_time": "",
            }
            self.history_state_list.append(history_state_item)
            self.previous_state = self.current_state
            logging.getLogger("HWR").debug(
                "StateMachine: current state " + "changed to : %s" % self.current_state
            )
            self.emit("stateChanged", self.history_state_list)

            self.update_fsm_state()

    def get_condition_list(self):
        """Returns list of conditions"""
        return self.condition_list

    def get_state_list(self):
        """Returns list of available state"""
        return self.state_list

    def get_transition_list(self):
        """Returns list of available transitions"""
        return self.transition_list

    def re_emit_values(self):
        """Reemits signals"""

        if len(self.history_state_list):
            self.emit("stateChanged", (self.history_state_list))



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

from enum import (
    Enum,
    unique,
)

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractNState

__credits__ = ["DESY P11"]
__license__ = "LGPLv3+"
__category__ = "General"

import time
from enum import (
    Enum,
    unique,
)

from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter


@unique
class BackLightValues(Enum):
    IN = "In"
    OUT = "Out"
    MOVING = "Moving"
    UNKNOWN = "UNKNOWN"


class P11BackLight(AbstractNState):
    """
    P11 BakcLight define interface to Tango backlight at DESY P11
    """

    VALUES = BackLightValues

    default_open_time = 8
    default_close_time = 3

    def __init__(self, name):
        super().__init__(name)

        self.cmd_open_close = None
        self.cmd_started = 0

        self.chan_state_open = None
        self.chan_state_close = None

    def init(self):
        """Initialise the predefined values"""

        self.chan_value = self.get_channel_object("value")
        self.open_time = self.get_property("open_time", self.default_open_time)
        self.close_time = self.get_property("close_time", self.default_close_time)

        if self.chan_value is not None:
            self.chan_value.connect_signal("update", self.update_light_state)

        self.update_light_state(self.chan_value.get_value())
        super().init()

    def get_value(self):
        return self.update_light_state()

    def is_moving(self):
        return self.get_value() == self.VALUES.MOVING

    def is_in(self):
        return self.get_value() == self.VALUES.IN

    def is_out(self):
        return self.get_value() == self.VALUES.OUT

    def set_in(self):
        self.set_value(self.VALUES.IN)

    def set_out(self):
        self.set_value(self.VALUES.OUT)

    def _set_value(self, value):
        current_value = self.chan_value.get_value()

        if value == self.VALUES.IN:
            new_value = 1
        elif value == self.VALUES.OUT:
            new_value = 0

        if current_value != new_value:
            self.chan_value.set_value(new_value)
            self.cmd_started = time.time()

    def update_light_state(self, value=None):
        """Updates light state

        :return: light state as str
        """
        if value is None:
            value = self.chan_value.get_value()

        elapsed = time.time() - self.cmd_started

        if value == 1:
            if elapsed < self.open_time:
                light_value = self.VALUES.MOVING
            else:
                light_value = self.VALUES.IN
        elif value == 0:
            if elapsed < self.close_time:
                light_value = self.VALUES.MOVING
            else:
                light_value = self.VALUES.OUT
        else:
            light_value = self.VALUES.UNKNOWN

        self.update_value(light_value)

        return light_value



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"
__author__ = "Jan Meyer"
__email__ = "jan.meyer@desy.de"


import logging

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractMotor import (
    AbstractMotor,
    MotorStates,
)


class DigitalZoomMotor(AbstractMotor, HardwareObject):
    """
    Works with camera devices which provide
    zoom_exists, set_zoom, get_zoom and get_zoom_min_max
    <object class="DigitalZoomMotor">
        <username>Zoom</username>
        <actuator_name>Zoom</actuator_name>
        <object href="/mjpg-stream-video" role="camera"/>
    </object>
    """

    def __init__(self, name):
        AbstractMotor.__init__(self, name)
        super().__init__(name)
        self.camera = None

    def init(self):
        self.set_limits((1.0, 1.0))
        try:
            self.camera = self.get_object_by_role("camera")
        except KeyError:
            logging.getLogger("HWR").warning("DigitalZoomMotor: camera not defined")
            return
        try:
            self.read_only = not (self.camera.zoom_exists())
        except AttributeError:
            self.read_only = True
        if not self.read_only:
            limits = self.camera.get_zoom_min_max()
            if limits[0] is None:
                limits[0] = -1000
            if limits[1] is None:
                limits[1] = 1000
            self.set_limits(limits)
            self.set_position(self.camera.get_zoom())
            self.update_state(self.STATES.READY)
        else:
            self.update_state(self.STATES.OFF)
            logging.getLogger("HWR").warning(
                "DigitalZoomMotor: digital zoom is not supported by camera object"
            )

    def update_state(self):
        """
        Descript. : forces position update
        """
        self.motor_position_changed()

    def motor_position_changed(self, position=None):
        """
        Descript. : called by move and updateState. if the position has
                    changed valueChanged is fired if the position is at one
                    of the limits the state is set accordingly on state
                    changes, stateChanged is fired
        """
        if position is None:
            if self.read_only:
                position = 1.0
            else:
                position = self.camera.get_zoom()

        if position != self.get_value():
            current_motor_state = self.get_state()
            if position <= self.get_limits()[0]:
                self.update_state(MotorStates.LOWLIMIT)
            elif position >= self.get_limits()[1]:
                self.update_state(MotorStates.HIGHLIMIT)
            else:
                current_motor_state = MotorStates.READY
            if (not self.read_only) and current_motor_state != self.get_state():
                self.update_state(current_motor_state)
                self.emit("stateChanged", (current_motor_state,))
            self.set_position(position)
            self.emit("valueChanged", (position,))

    #    def get_limits(self):
    #        """
    #        Descript. : returns motor limits. If no limits channel defined then
    #                    static_limits is returned
    #        """
    #        return self.limits

    def get_value(self):
        if self.read_only:
            self.motor_position = 1.0
        else:
            self.motor_position = self.camera.get_zoom()
        return self.motor_position

    def _set_value(self, value):
        """
        Descript. : move to the given position
        """
        self.camera.set_zoom(value)
        self.motor_position_changed(value)

    def stop(self):
        """
        Descript. : does nothing, for position change is instantaneous
        """

    def is_moving(self):
        return False



import logging
import time

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.CatsMaint import CatsMaint


class PX1CatsMaint(CatsMaint):
    def __init__(self, *args):
        CatsMaint.__init__(self, *args)
        self.home_opened = None

    def init(self):
        CatsMaint.init(self)

        self._chnHomeOpened = self.add_channel(
            {
                "type": "tango",
                "name": "_chnHomeOpened",
                "tangoname": self.tangoname,
                "polling": "events",
            },
            "homeOpened",
        )

        self._chnHomeOpened.connect_signal("update", self.update_home_opened)

        self._cmdDrySoak = self.add_command(
            {"type": "tango", "name": "_cmdDrySoak", "tangoname": self.tangoname},
            "DryAndSoak",
        )

        self._cmdReset = self.add_command(
            {"type": "tango", "name": "_cmdReset", "tangoname": self.tangoname},
            "ResetError",
        )

    def update_home_opened(self, value):
        if value != self.home_opened:
            self.home_opened = value
            self._update_global_state()

    def get_global_state(self):
        state_dict, cmd_state, message = CatsMaint.get_global_state(self)
        state_dict["homeopen"] = self.home_opened
        return state_dict, cmd_state, message

    def _do_home_open(self, unload=False):
        if unload and self.loaded:
            logging.getLogger("HWR").debug("Unloading sample first")
            self.cats_hwo._do_unload()
            time.sleep(3)
            while HWR.beamline.sample_changer._is_device_busy():
                time.sleep(0.3)

        logging.getLogger("HWR").debug("Running the home command (home/open) now")
        self._cmdHome()

    def _do_dry_soak(self):
        self._cmdDrySoak()

    def _do_reset(self):
        logging.getLogger("HWR").debug("PX1CatsMaint: executing the _do_reset function")
        self._cmdReset()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Actuator."""

from __future__ import annotations

import abc
import math
from ast import literal_eval

from gevent.lock import RLock

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractActuator(HardwareObject):
    """Abstract actuator defines methods common to all moving devices.

    The ``_set_value`` method is the only abstract method that needs to be overloaded
    in each implementation.

    Attributes:
        _nominal_value (float | None):
            Current actuator value.
        default_value (float | None):
            Value specified by XML property, otherwise ``None``.
        _nominal_limits (tuple[float | None, float | None]):
            Values specified by XML property, otherwise ``None``.
        actuator_name (str | None):
            Actuator name specified by XML property, otherwise ``None``.
        read_only (bool):
            Read-only flag specified by XML property, otherwise ``False``.
        username (str):

    Emits:
        valueChanged (tuple[int]):
            Tuple whose first and only item is the new value.
            Emitted during initialization of the hardware object
            and when setting a new value.
        limitsChanged (tuple[tuple[int, int]]):
            Tuple whose first and only item is a two-item tuple of the new limits
            (low limit first and high limit second).
            Emitted by ``update_limits`` if limit values are changed.
        stateChanged (tuple):
            Tuple whose first and only item is the new state.
            Emitted by ``force_emit_signals``
    """

    __metaclass__ = abc.ABCMeta

    unit = None

    def __init__(self, name: str):
        super().__init__(name)
        self._nominal_value = None
        self._nominal_limits = (None, None)
        self.actuator_name = None
        self.read_only = False
        self.default_value = None
        self.username = None
        self._lock = RLock()

    def init(self):
        """Init properties: actuator_name, username, read_only and default_value."""
        self.actuator_name = self.get_property("actuator_name")
        self.read_only = self.get_property("read_only") or False
        self.default_value = self.get_property("default_value")
        if self.default_value is not None:
            self.update_value(self.default_value)
        limits = self.get_property("default_limits")
        if limits:
            try:
                self._nominal_limits = tuple(literal_eval(limits))
            except TypeError:
                print("Invalid limits")
        self.username = self.get_property("username")

    @abc.abstractmethod
    def get_value(self):
        """Read the actuator position.

        Returns:
            Actuator position.
        """
        return None

    def get_limits(self):
        """Return actuator low and high limits.

        Returns:
            (tuple): Two-item tuple (low limit, high limit).
        """
        return self._nominal_limits

    def set_limits(self, limits: tuple) -> None:
        """Set actuator low and high limits and emit signal ``limitsChanged``.

        Args:
            limits (tuple): Two-item tuple (low limit, high limit).

        Raises:
            ValueError: Attempt to set limits for read-only actuator.
        """
        if self.read_only:
            raise ValueError("Attempt to set limits for read-only Actuator")

        self._nominal_limits = limits
        self.emit("limitsChanged", (self._nominal_limits,))

    def validate_value(self, value) -> bool:
        """Check if the value is within limits.

        Args:
            value(numerical): Value.

        Returns:
            ``True`` if within the limits, ``False`` otherwise.
        """
        if value is None:
            return True
        if math.isnan(value) or math.isinf(value):
            return False
        if None in self._nominal_limits:
            return True
        return self._nominal_limits[0] <= value <= self._nominal_limits[1]

    @abc.abstractmethod
    def _set_value(self, value):
        """Implementation of specific set actuator logic.

        Args:
            value: Target value.
        """

    def set_value(self, value, timeout: float = 0) -> None:
        """Set actuator to value.

        If ``timeout == 0``: return at once and do not wait (default).
        If ``timeout is None``: wait forever.

        Args:
            value: target value
            timeout (float): Optional timeout in seconds. Default is ``0``: do not wait.

        Raises:
            ValueError: Invalid value or attempt to set read only actuator.
            RuntimeError: Timeout waiting for status ready (from ``wait_ready``):
        """

        with self._lock:
            if self.read_only:
                raise ValueError("Attempt to set value for read-only Actuator")
            if self.validate_value(value):
                self._set_value(value)
                self.update_value()
                if timeout == 0:
                    return
                self.wait_ready(timeout)
            else:
                raise ValueError(f"Invalid value {value}")

    def update_value(self, value=None) -> None:
        """Check if the value has changed and emit signal ``valueChanged``.

        Args:
            value: Value.
        """
        if value is None:
            value = self.get_value()

        if self._nominal_value != value:
            self._nominal_value = value
            self.emit("valueChanged", (value,))

    def update_limits(self, limits=None) -> None:
        """Check if the limits have changed and emit signal ``limitsChanged``.

        Args:
            limits (tuple): Two-item tuple (low limit, high limit).
        """
        if not limits:
            limits = self.get_limits()

        if self._nominal_limits != limits:
            # All values are not NaN
            if not any(isinstance(lim, float) and math.isnan(lim) for lim in limits):
                self._nominal_limits = limits
                self.emit("limitsChanged", (limits,))

    def re_emit_values(self) -> None:
        """Update values for all internal attributes."""
        self.update_value(self.get_value())
        self.update_limits(self.get_limits())
        super(AbstractActuator, self).re_emit_values()

    def force_emit_signals(self) -> None:
        """Force emission of all signals.

        Method is called from GUI.
        Do not call it from within a hardware object.
        """
        self.emit("valueChanged", (self.get_value(),))
        self.emit("limitsChanged", (self.get_limits(),))
        self.emit("stateChanged", (self.get_state(),))
